# -*- coding: utf-8 -*-
import locale
from datetime import datetime, timedelta
from time import sleep
import curses
import logging
import threading
import platform

from .simple_curses_widgets import SimpleCursesCheckBox, SimpleCursesPushButton, SimpleCursesTime, SimpleCursesString, SimpleCursesDate, SimpleCursesLineEdit
from .schedule import PyRadioScheduleItem, PyRadioScheduleItemType, PyRadioScheduleTimeType, PyRadioTime, PyRadioScheduleList, format_date_to_iso8851, datetime_to_my_time, is_date_before
from .keyboard import kbkey, check_localized

locale.setlocale(locale.LC_ALL, '')    # set your locale

logger = logging.getLogger(__name__)

class PyRadioSimpleScheduleWindow():
    _error_messages = {
        3: '_______Entry is |invalid|!\n__|No |Start| or |Stop| time specified!',
        6: '___Entry is |invalid|!\n__Start time| is in the |past|!',
        7: '___Entry is |invalid|!\n__End time| is in the |past|!',
        8: '______Entry is |invalid|!\n__Start time| is after |End time|!'
    }

    _repeat = (
        'day',
        'week',
        'month',
        'Monday',
        'Tuesday',
        'Wednesday',
        'Thursday',
        'Friday',
        'Saturday',
        'Sunday'
    )

    '''     0  1  2   3  4  5   6   7  8  9  10 11 12 13 14  15  16  17  18  19  20  21  22 '''
    _up = (22, 0, 1, 19, 2, 3, 20, 21, 4, 5, 8, 9, 6, 7, 10, 14, 11, 15, 17, 16, 12, 13, 18)

    _tips = (
        'The playlist to open',
        'The stations to play',
        'Enable Start Timer (time to start playback)',
        'The date the playback will start',
        'Absolute time to start playback',
        'Absolute time to start playback',
        'Start playback XX:XX:XX after the time "OK" is pressed',
        'Start playback XX:XX:XX after the time "OK" is pressed',
        'Enable Stop Timer (time to stop playback)',
        'The date the playback will stop',
        'Absolute time to stop playback',
        'Absolute time to stop playback',
        'Stop playback XX:XX:XX after the "Start" time',
        'Stop playback XX:XX:XX after the "Start" time',
        'The player to use',
        'Enable recording the station to a file',
        'Play no sound while recording',
        'Enable buffering (default parameters will be used)',
        'Enable recurrence for this schedule entry',
        'Recurrence parameter for this schedule entry',
        'Save changes to this schedule schedule entry',
        'Cancel changes to this schedule entry',
        'A descriptive name for this schedule entry'
    )

    def __init__(
            self, parent,
            supported_players,
            current_player,
            my_op_mode,
            cur_op_mode,
            playlist=None, station=None,
            schedule_item=None,
            global_functions=None
    ):
        self._info_result = None
        self._X = 0
        self._Y = 0
        self._maxY = 0
        self.maxX = 0
        self._widgets = None
        self._win = None
        self._global_functions = {}
        self._showed = False
        self._too_small = False
        self._stop = False
        self._thread_date = None
        self._error_num = 0
        self._max_repeat_len = max(len(x) for x in self._repeat)
        self._repeat_index = 0
        self.lock = threading.Lock()
        self._entry = None
        self._exit = False
        self._my_op_mode = my_op_mode
        self._cur_op_mode = cur_op_mode
        self._playlist = playlist
        self._station = station
        self._supported_players = supported_players
        self._maxX = 60
        self._displacement = 3
        self._maxY = 20
        if global_functions is None:
            self._global_functions = {}
        else:
            self._global_functions = global_functions
        self._get_parent(parent)
        self._focus = 22
        try:
            self._current_player_id = supported_players.index(current_player)
        except ValueError:
            self._current_player_id = 0
        self._player_id = self._current_player_id
        self.set_item(schedule_item, playlist, station)

    def __del__(self):
        self._exit = True
        self._stop = True

    def exit(self):
        self._stop = True
        self._exit = True

    @property
    def entry(self):
        return self._entry

    @property
    def too_small(self):
        return self._too_small

    @property
    def playlist(self):
        return self._schedule_item.item['playlist']

    @playlist.setter
    def playlist(self, val):
        self._schedule_item.item['playlist'] = val
        self._widgets[0].string = val

    @property
    def station(self):
        return self._schedule_item.item['station']

    @station.setter
    def station(self, val):
        self._schedule_item.item['station'] = val
        self._widgets[1].string = val

    def get_error_message(self):
        return self._error_messages[self._error_num]

    def set_item(self, schedule_item=None, playlist=None, station=None):
        if schedule_item is None:
            self._schedule_item = PyRadioScheduleItem()
            # self._schedule_item.item['type'] = 0
            # self._schedule_item.item['end_date'] = [2023, 11, 12]
            # self._schedule_item.item['end_time'] = [20, 1, 2, 0]
            # self._schedule_item.item['start_duration'] = [1, 32, 15, 0]
            # self._schedule_item.item['end_duration'] = [3, 15, 2, 0]
            # self._schedule_item.item['repeat'] = 'Sunday'
            if playlist:
                self._schedule_item.playlist = playlist
            if station:
                self._schedule_item.station = station
            self._remove_enabled = False
        else:
            self._schedule_item = PyRadioScheduleItem(schedule_item)
            self._remove_enabled = True

        logger.error('set_item')
        for n, item in self._schedule_item.item.items():
            logger.error(f'{n}: {item}')

        ''' parse and assign values '''

    def _move_widgets(self):
        # self._widgets[0].move(1,2)
        # self._widgets[1].move(2,2)
        # Start
        self._widgets[2].move(6 + self._Y, 5 + self._X)
        self._widgets[3].move(6, self._widgets[2].X + self._widgets[2].width - self._X)
        self._widgets[4].move(7 + self._Y, 9 + self._X)
        self._widgets[5].move(7, self._widgets[4].X + self._widgets[4].width - self._X)
        self._widgets[6].move(7 + self._Y, self._X + self._widgets[5].X + self._widgets[5].width + 4)
        self._widgets[7].move(7, self._widgets[6].X + self._widgets[6].width - self._X)
        # End
        self._widgets[8].move(8 + self._Y, 5 + self._X)
        self._widgets[9].move(8, self._widgets[8].X + self._widgets[8].width - self._X)
        self._widgets[10].move(9 + self._Y, 9 + self._X)
        self._widgets[11].move(9, self._widgets[10].X + self._widgets[10].width - self._X)
        self._widgets[12].move(9 + self._Y, self._X + self._widgets[11].X + self._widgets[11].width + 4)
        self._widgets[13].move(9, self._widgets[12].X + self._widgets[12].width - self._X)
        # options
        # self._widgets[14].move(8, 3)
        self._widgets[15].move(9 + self._displacement + self._Y, self._X + 9)
        self._widgets[16].move(9 + self._displacement + self._Y, self._widgets[15].X + self._widgets[15].width + 4)
        self._widgets[17].move(10 + self._displacement + self._Y, self._X + 9)
        self._widgets[18].move(11 + self._displacement + self._Y, self._X + 5)
        # self._widgets[19].move(8 + self._displacement + self._Y, self._widgets[17].X + self._widgets[18].width)
        # Buttons
        self._widgets[20].move(12 + self._displacement + self._Y, self._X + self._maxX - 19)
        self._widgets[21].move(self._widgets[20].Y, self._X + self._maxX - (len(self._widgets[13].caption) + 12))


        # self._widgets[21].move(12 + self._displacement + self._Y, self._X + self.maxX - self._widgets[21].width)
        self._widgets[22].move(
                self._win,
                self._Y + self._widgets[0].Y - 1,
                self._X + 12,
                )

    def _get_parent(self, parent):
        self._parent = parent
        self._parent_maxY, self._parent_maxX = parent.getmaxyx()
        new_Y = int((self._parent_maxY - self._maxY) / 2) + 1
        new_X = int((self._parent_maxX - self._maxX) / 2)
        if self._Y != new_Y or self._X != new_X:
            self._Y = new_Y
            self._X = new_X
        self._win = curses.newwin(self._maxY, self._maxX, self._Y, self._X)
        self._init_win()

    def _init_win(self):
        self._win.bkgdset(' ', curses.color_pair(3))
        self._win.erase()
        self._win.box()
        self._print_header()

    def _print_header(self):
        msg = ' Schedule editor '
        self._win.addstr(
            0, int((self._maxX - len(msg)) / 2),
            msg, curses.color_pair(11)
        )

    def move(self, new_Y=None, new_X=None):
        if new_Y:
            self._Y = new_Y
        if new_X:
            self._X = new_X
        if self._win:
            self._win.mvwin(self._Y, self._X)

    def _add_widgets(self):
        self._widgets = []

        ''' id 0 playlist string '''
        self._widgets.append(
            SimpleCursesString(
                Y=2,X=2,
                parent=self._win,
                caption='Playlist: ',
                string='Any playlist',
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                color_not_focused=curses.color_pair(11),
                color_disabled=curses.color_pair(10),
                max_selection = 46
            )
        )
        self._widgets[-1].w_id = 0

        ''' id 1 stations string '''
        self._widgets.append(
            SimpleCursesString(
                Y=3, X=2,
                parent=self._win,
                caption=' Station: ',
                string='Any station',
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                color_not_focused=curses.color_pair(11),
                color_disabled=curses.color_pair(10),
                max_selection = 46
            )
        )
        self._widgets[-1].w_id = 1

        ''' id 2 start check box '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'Start playback on: ',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 2

        ''' id 3 start at date box '''
        self._widgets.append(
            SimpleCursesDate(
                Y=0, X=0,
                window=self._win,
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                next_widget_func=self._next_widget,
                previous_widget_func=self._previous_widget,
                global_functions=self._global_functions
            )
        )
        self._widgets[-1].w_id = 3

        ''' id 4 start at check box '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'At: ',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 4

        ''' id 5 start at time '''
        self._widgets.append(
            SimpleCursesTime(
                Y=7, X=0,
                window=self._win,
                color=curses.color_pair(10),
                show_am_pm=True,
                color_focused=curses.color_pair(9),
                lock=self.lock,
                next_widget_func=self._next_widget,
                previous_widget_func=self._previous_widget,
                global_functions=self._global_functions,
                string=PyRadioTime.pyradio_time_to_string(self._schedule_item.item['start_time'])
            )
        )
        self._widgets[-1].w_id = 5

        ''' id 6 start in check box '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'In: ',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 6

        ''' id 7 start in time '''
        self._widgets.append(
            SimpleCursesTime(
                Y=7, X=0,
                window=self._win,
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                lock=self.lock,
                next_widget_func=self._next_widget,
                previous_widget_func=self._previous_widget,
                global_functions=self._global_functions,
                string=PyRadioTime.pyradio_time_to_string(self._schedule_item.item['start_duration'])
            )
        )
        self._widgets[-1].w_id = 7

        ''' id 8 stop check box '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0,0, 'End playback on: ',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 8

        ''' id 9 stop at date box '''
        self._widgets.append(
            SimpleCursesDate(
                Y=7, X=0,
                window=self._win,
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                next_widget_func=self._next_widget,
                previous_widget_func=self._previous_widget,
                global_functions=self._global_functions
            )
        )
        self._widgets[-1].w_id = 9

        ''' id 10 stop at check box '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'At: ',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 10

        ''' id 11 stop at time '''
        self._widgets.append(
            SimpleCursesTime(
                Y=9, X=0,
                window=self._win,
                show_am_pm=True,
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                lock=self.lock,
                next_widget_func=self._next_widget,
                previous_widget_func=self._previous_widget,
                global_functions=self._global_functions,
                string=PyRadioTime.pyradio_time_to_string(self._schedule_item.item['end_time'])
            )
        )
        self._widgets[-1].w_id = 11

        ''' id 12 stop in check box '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'In: ',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 12

        ''' id 13 stop in time '''
        self._widgets.append(
            SimpleCursesTime(
                Y=9,X=0,
                window=self._win,
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                lock=self.lock,
                next_widget_func=self._next_widget,
                previous_widget_func=self._previous_widget,
                global_functions=self._global_functions,
                string=PyRadioTime.pyradio_time_to_string(self._schedule_item.item['end_duration'])
            )
        )
        self._widgets[-1].w_id = 13


        ''' id 14 player '''
        self._widgets.append(
            SimpleCursesString(
                Y=11, X=3,
                parent=self._win,
                caption='Player: ',
                string=self._supported_players[self._current_player_id],
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                color_not_focused=curses.color_pair(11),
                color_disabled=curses.color_pair(10),
                max_selection = 10
            )
        )
        self._widgets[-1].w_id = 14

        ''' id 15 recording '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'Record to file',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 15

        ''' id 16 silent recording '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'Silent recording',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 16

        ''' id 17 buffering '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'Use buffering',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 17

        ''' id 18 cancel scheduling button '''
        self._widgets.append(
            SimpleCursesCheckBox(
                0, 0, 'Repeat every:',
                curses.color_pair(9), curses.color_pair(11), curses.color_pair(10)
            )
        )
        self._widgets[-1].w_id = 18

        ''' id 19 silent recording '''
        self._widgets.append(
            SimpleCursesString(
                Y=14,
                X=23,
                parent=self._win,
                caption='',
                string='day',
                color=curses.color_pair(10),
                color_focused=curses.color_pair(9),
                color_not_focused=curses.color_pair(11),
                color_disabled=curses.color_pair(10),
                max_selection = 10
            )
        )
        self._widgets[-1].w_id = 19

        ''' id 20 ok button '''
        cap = 'OK'
        self._widgets.append(SimpleCursesPushButton(
            Y=0, X=0,
            caption=cap,
            color_focused=curses.color_pair(9),
            color=curses.color_pair(11),
            bracket_color=curses.color_pair(10),
            parent=self._parent))
        self._widgets[-1].w_id = 20

        ''' id 21 cancel button '''
        cap = 'Cancel'
        self._widgets.append(SimpleCursesPushButton(
            Y=0, X=0,
            caption=cap,
            color_focused=curses.color_pair(9),
            color=curses.color_pair(11),
            bracket_color=curses.color_pair(10),
            parent=self._parent))
        self._widgets[-1].w_id = 21

        ''' id 22 name line editor '''
        self._widgets.append(SimpleCursesLineEdit(
            parent=self._win,
            width=46,
            begin_y=self._Y + self._widgets[0].Y - 1,
            begin_x=self._X + 12,
            boxed=False,
            has_history=False,
            caption='',
            box_color=curses.color_pair(9),
            caption_color=curses.color_pair(11),
            edit_color=curses.color_pair(9),
            cursor_color=curses.color_pair(8),
            unfocused_color=curses.color_pair(11),
            key_up_function_handler=self._go_up,
            key_down_function_handler=self._go_down,
            key_tab_function_handler=self._next_widget,
            key_stab_function_handler=self._previous_widget
            )
        )
        ''' enables direct insersion of ? and \\ '''
        self._widgets[-1]._paste_mode = False
        self._widgets[-1].w_id = 22
        self._widgets[-1].string = self._schedule_item.item['name']

        self._move_widgets()

    def show(self, parent=None):
        if parent:
            self._get_parent(parent)
            if self._showed:
                self._move_widgets()
        if self._widgets is None:
            self._add_widgets()
        with self.lock:
            try:
                self._win.addstr(self._maxY-3, 2, '─' * (self._maxX - 4), curses.color_pair(3))
            except:
                self._win.addstr(self._maxY-3, 2, '─'.encode('utf-8') * (self._maxX - 4), curses.color_pair(3))
            self._win.addstr(self._maxY-3, 3, ' Tip ', curses.color_pair(11))
        disp_msg = self._tips[self._focus]
        with self.lock:
            self._win.addstr(1, 6, 'Name: ',  curses.color_pair(10))
            self._win.addstr(self._maxY-2, 2, disp_msg.ljust(self._maxX-4), curses.color_pair(10))
            self._win.refresh()

        if self._widgets:
            if self._showed:
                self._fix_focus()
            else:
                self._item_to_form()
            for n, _ in enumerate(self._widgets):
                if n == len(self._widgets) -1:
                    with self.lock:
                        self._widgets[n].show(parent_win=self._win, opening=False)
                else:
                    try:
                        with self.lock:
                            self._widgets[n].show(self._win)
                    except:
                        logger.error('exception!')
                        with self.lock:
                            self._widgets[n].show()
        with self.lock:
            self._win.refresh()
        self._showed = True
        if self._thread_date is None:
            self._thread_date = threading.Thread(
                target=self._thread_show_date,
                args =(
                    lambda: self._win,
                    self.lock,
                    lambda: self._can_show_date,
                    lambda: self._stop
                )
            )
            self._thread_date.start()

    def _can_show_date(self):
        if self._too_small:
            return False
        if self._my_op_mode == self._cur_op_mode():
            return True
        return False

    def _apply_repeat_to_widget(self):
        repeat = self._schedule_item.item['repeat']
        if repeat is None:
            self._repeat_index = 0
        else:
            if repeat in self._repeat:
                self._repeat_index = self._repeat.index(repeat)
            else:
                self._repeat_index = 0
        self._widgets[19].string = self._repeat[self._repeat_index].ljust(self._max_repeat_len)

    def _item_to_form(self):
        logger.error(f'\n\n{self._schedule_item.item}\n\n')
        ''' assign values '''
        self._widgets[3].date = self._schedule_item.item['start_date']
        self._widgets[5].time = self._schedule_item.item['start_time']
        self._widgets[7].time = self._schedule_item.item['start_duration']
        self._widgets[9].date = self._schedule_item.item['end_date']
        self._widgets[11].time = self._schedule_item.item['end_time']
        self._widgets[13].time = self._schedule_item.item['end_duration']

        if self._schedule_item.item['buffering'] == 0:
            self._widgets[17].checked = False
        else:
            self._widgets[17].checked = True

        if self._schedule_item.item['type'] == PyRadioScheduleItemType.TYPE_END:
            self._widgets[0].string = 'Any playlist'
            self._widgets[1].string = 'Any station'
            self._widgets[0].enabled = False
            self._widgets[1].enabled = False
            rec = False
            start_fields = False
            end_fields = True
        else:
            self._widgets[0].string = self._schedule_item.item['playlist']
            self._widgets[1].string = self._schedule_item.item['station']
            self._widgets[0].enabled = True
            self._widgets[1].enabled = True
            rec = True
            start_fields = True
            end_fields = True

        self._widgets[18].enabled = False
        if self._schedule_item.item['type'] == PyRadioScheduleItemType.TYPE_START_END:
            self._widgets[2].checked = self._widgets[8].checked = True
            self._widgets[18].enabled = True
            logger.error(f'1 setting {self._widgets[18].enabled = }')
        elif self._schedule_item.item['type'] == PyRadioScheduleItemType.TYPE_START:
            self._widgets[2].checked = True
            self._widgets[8].checked = False
            self._widgets[18].enabled = True
            logger.error(f'2 setting {self._widgets[18].enabled = }')
        elif self._schedule_item.item['type'] == PyRadioScheduleItemType.TYPE_END:
            self._widgets[2].checked = False
            self._widgets[8].checked = True
            self._widgets[18].enabled = True
            logger.error(f'3 setting {self._widgets[18].enabled = }')


        if self._schedule_item.item['start_type'] == PyRadioScheduleTimeType.TIME_ABSOLUTE:
            self._widgets[4].checked = self._widgets[5].enabled = True
            self._widgets[6].checked = self._widgets[7].enabled = False
        else:
            self._widgets[4].checked = self._widgets[5].enabled = False
            self._widgets[6].checked = self._widgets[7].enabled = True

        if self._schedule_item.item['end_type'] == PyRadioScheduleTimeType.TIME_ABSOLUTE:
            self._widgets[10].checked = self._widgets[11].enabled = True
            self._widgets[12].checked = self._widgets[13].enabled = False
        else:
            self._widgets[10].checked = self._widgets[11].enabled = False
            self._widgets[12].checked = self._widgets[13].enabled = True

        self._fix_focus()

    def _fix_focus(self):
        for i in (2, 8):
            for k in range(2, 6):
                logger.error(f'1 setting {i + k} enabled to {self._widgets[i].checked}')
                self._widgets[i+k].enabled = self._widgets[i].checked

        for i in range(14, 18):
            # logger.debug('{0} enabled is now {1}'.format(i, self._widgets[2].checked))
            logger.error(f'2 setting {i} enabled to {self._widgets[2].checked}')
            self._widgets[i].enabled = self._widgets[2].checked

        # self._widgets[-2].enabled = self._widgets[2].checked or self._widgets[8].checked

        for i in (4, 6, 10, 12):
            if self._widgets[i].enabled:
                logger.error(f'3 setting {i} enabled to {i} - {self._widgets[i].checked}')
                self._widgets[i+1].enabled = self._widgets[i].checked

        # for i in (6, 12):
        #     if self._widgets[i].checked:
        #         logger.error('setting {} enabled to False'.format(i-3))
        #         self._widgets[i-3].enabled = False
        #     else:
        #         logger.error('setting {} enabled to True'.format(i-3))
        #         self._widgets[i-3].enabled = True

        for i in (15, 18):
            if self._widgets[i].enabled:
                logger.error(f'4 setting {i + 1} enabled to {self._widgets[i].checked}')
                self._widgets[i+1].enabled = self._widgets[i].checked

        #fix repeat field
        if self._widgets[2].checked or self._widgets[8].checked:
            self._widgets[18].enabled = True
        else:
            self._widgets[18].enabled = False

        if self._widgets[18].checked and self._widgets[18].enabled:
            # repeat is on
            if self._widgets[2].checked:
                # start will be absolute
                self._widgets[3].enabled = False
                self._widgets[4].enabled = True
                self._widgets[4].checked = True
                self._widgets[5].enabled = True
                self._widgets[6].checked = False
                self._widgets[6].enabled = False
                self._widgets[7].enabled = False
            else:
                if self._widgets[8].checked:
                    # end will be absolute if start not checked
                    # this is stop-only
                    self._widgets[9].enabled = False
                    self._widgets[10].enabled = True
                    self._widgets[10].checked = True
                    self._widgets[11].enabled = True
                    self._widgets[12].checked = False
                    self._widgets[12].enabled = False
                    self._widgets[13].enabled = False
        else:
            # repeat is off
            self._widgets[3].enabled = self._widgets[2].checked
            self._widgets[9].enabled = self._widgets[8].checked
            self._widgets[6].enabled = True
            self._widgets[12].enabled = True

        self._apply_repeat_to_widget()

        if platform.system().lower().startswith('win') and \
                self._widgets[14].string == 'vlc':
            self._widgets[15].checked = False
            self._widgets[15].enabled = False
            self._widgets[16].enabled = False
        else:
            if self._schedule_item.item['recording'] == 1:
                self._widgets[15].checked = True
                self._widgets[16].checked = False
                self._widgets[16].enabled = True
            elif self._schedule_item.item['recording'] == 2:
                self._widgets[15].checked = True
                self._widgets[16].checked =  True
                self._widgets[16].enabled = True
            else:
                self._widgets[15].checked = False
                self._widgets[16].enabled = False
                self._widgets[16].checked = False

        self._widgets[20].enabled = self._widgets[2].checked or self._widgets[8].checked

        for i, _ in enumerate(self._widgets):
            if self._focus == i:
                self._widgets[i].focused = True
            else:
                self._widgets[i].focused = False

    def _go_up(self):
        self._focus = self._up[self._focus]
        while not self._widgets[self._focus].enabled:
            self._focus = self._up[self._focus]
        if self._focus in (5, 7, 11, 13):
            self._widgets[self._focus].reset_selection()

    def _go_down(self):
        self._focus = self._up.index(self._focus)
        while not self._widgets[self._focus].enabled:
            self._focus = self._up.index(self._focus)
        if self._focus in (5, 7, 11, 13):
            self._widgets[self._focus].reset_selection()

    def _next_widget(self):
        old_focus = self._focus
        self._focus += 1
        if self._focus >= len(self._widgets):
            self._focus = 0

        while not self._widgets[self._focus].enabled:
            self._focus += 1

        if self._focus in (5, 7, 11, 13):
            self._widgets[self._focus].reset_selection()
        self.show()

    def _previous_widget(self):
        self._focus -= 1
        if self._focus < 0:
            self._focus = len(self._widgets) - 1

        while not self._widgets[self._focus].enabled:
            self._focus -= 1
            if self._focus == -1:
                self._focus = len(self._widgets) -1

        if self._focus in (5, 7, 11, 13):
            self._widgets[self._focus].reset_selection(last=True)
        self.show()

    def _form_to_item(self):
        rec = 1 if self._widgets[15].checked else 0
        rep = None
        if self._widgets[18].checked:
            rep = self._widgets[19].string.strip()
        if rec == 1 and self._widgets[16].checked:
            rec = 2
        if self._widgets[2].checked:
            ''' start and stop '''
            playlist = self._widgets[0].string
            station = self._widgets[1].string
            if self._widgets[8].checked:
                the_type = 0
            else:
                the_type = 1

            buf = 1 if self._widgets[17].checked else 0
        else:
            ''' stop only '''
            the_type = 2
            rep = None
            buf = 0
            rec = 0
            playlist = None
            station = None
        tmp_item = PyRadioScheduleItem({
            'name': self._widgets[22].string,
            'type': the_type, # TYPE_START_END, TYPE_START, TYPE_END
            'start_type': 0 if self._widgets[4].checked else 1, # TIME_ABSOLUTE, TIME_RELATIVE
            'start_date':  list(self._widgets[3].date_tuple),
            'start_time': self._widgets[5].get_time()[0], # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'start_duration': self._widgets[7].get_time()[0], # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'end_type': 0 if self._widgets[10].checked else 1, # TIME_ABSOLUTE, TIME_RELATIVE
            'end_date': list(self._widgets[9].date_tuple),
            'end_time': self._widgets[11].get_time()[0], # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'end_duration': self._widgets[13].get_time()[0], # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'player': self._widgets[14].string,
            'recording': rec,
            'buffering': buf,
            'repeat': rep,
            'playlist': playlist,
            'station': station,
            'token': self._schedule_item.item['token']
        })
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('--== item ==--')
            for n, item in tmp_item.item.items():
                logger.debug(f'{n}: {item}')
        return tmp_item

    def _show_info(self):
        ret = 10
        if not self._widgets[2].checked and not self._widgets[8].checked:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Invalid entry')
                self._info_result = '''
                    _______--== |Invalid entry| ==--
                    ___No |Start| or |Stop| time specified___
                '''
                return 10
        tmp_item = self._form_to_item()
        ac_tmp_item = tmp_item.get_active_item()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('--== active item ==--')
            for n, item in tmp_item.active_item.items():
                logger.debug(f'{n}: tmp_item.active_{item}')
        ac_tmp_item = tmp_item.get_active_item()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'ac_tmp_item = {ac_tmp_item}')

        err_out = []
        error = False
        if tmp_item.item['repeat'] is None:
            if tmp_item.item['type'] == 0:
                if is_date_before(ac_tmp_item[2], ac_tmp_item[1]):
                    if not error:
                        err_out.append('--== |Invalid entry| ==--')
                        error = True
                    err_out.append('|Start time| is after |End time|!')
                elif is_date_before(ac_tmp_item[1], datetime.now()):
                    if not error:
                        err_out.append('--== |Invalid entry| ==--')
                        error = True
                    err_out.append('|Start time| is in the |past|!')
                    error = True
                    if is_date_before(ac_tmp_item[2], datetime.now()):
                        if not error:
                            err_out.append('--== |Invalid entry| ==--')
                            error = True
                        err_out.append('|Stop time| is in the |past|!')
            elif tmp_item.item['type'] == 1:
                if is_date_before(ac_tmp_item[1], datetime.now()):
                    err_out.append('--== |Invalid entry| ==--')
                    err_out.append('|Start time| is in the |past|!')
                    error = True
            elif tmp_item.item['type'] == 2:
                if is_date_before(ac_tmp_item[2], datetime.now()):
                    err_out.append('--== |Invalid item| ==--')
                    err_out.append('|Stop time| is in the |past|!')
                    error = True
        else:
            if tmp_item.item['type'] == 0:
                if is_date_before(ac_tmp_item[2], ac_tmp_item[1]):
                    err_out.append('--== |Invalid entry| ==--')
                    err_out.append('|Start time| is after |End time|!')
                    error = True
        out = []
        if tmp_item.item['repeat'] is None or error:
            out.extend(self._format_info_lines(tmp_item.item['type'], ac_tmp_item))
        else:
            # get list of occurrences
            it_count = 3 if tmp_item.item['type'] == 0 else 6
            the_r_list = PyRadioScheduleList(a_file='', a_list=[tmp_item.item])
            the_list = the_r_list.get_list_of_repeating_dates(it_count)
            out.append(f'Displaying |{it_count}| subsequent occurrences:')
            for n in the_list:
                out.append('__|#| Start: |' + n[0] + '__')
                if tmp_item.item['type'] == 0:
                    out.append('_____Stop: |' + n[1] + '__')

        if err_out:
            length = max(len(x) for x in out)
            for i, item in enumerate(err_out):
                num = int((length - len(item.replace('|', ''))) / 2)
                err_out[i] = num * '_' + item
            err_out.append('')
            out = err_out + out

        while max(len(x) for x in out) < 40:
            for i, _ in enumerate(out):
                out[i] = '__' + out[i] + '__'

        if out[0] != '':
            out.reverse()
            out.append('')
            out.reverse()
        if out[-1] != '':
            out.append('')
        self._info_result = '\n'.join(out)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'self._info_result\n{self._info_result}')
            logger.debug(f'ret = {ret}')

    def _format_info_lines(self, the_type, item):
        out = []
        if the_type == 0:
            out.append('__# Start: |' + datetime_to_my_time(item[1]) + '__')
            out.append('_____Stop: |' + datetime_to_my_time(item[2]) + '__')
        elif the_type == 1:
            out.append('__# Start: |' + datetime_to_my_time(item[1]) + '__')
        elif the_type == 2:
            out.append('__#__Stop: |' + datetime_to_my_time(item[2]) + '__')
        return out

    def _validate_selection(self):
        ''' validate form input
            Return
              0: All ok
                 Entry is set to self.entry
              3: Invalid entry
              6: Start time is in the past
              7: Stop time is in the past
              8: Stop time before Start time
        '''
        self._entry = None
        if not self._widgets[2].checked and not self._widgets[8].checked:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Invalid entry')
                return 3
        tmp_item = self._form_to_item()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('--== active item ==--')
            for n, item in tmp_item.active_item.items():
                logger.debug(f'{n}: tmp_item.active_{item}')
            logger.debug(f'tmp_item ={tmp_item}')
        ac_tmp_item = tmp_item.get_active_item()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'ac_tmp_item = {ac_tmp_item}')

        the_type = ac_tmp_item[3]
        rep = ac_tmp_item[-4]

        logger.error(f'\nthe_type = {the_type}')
        logger.error(f'rep = {rep}\n')

        if the_type in (0, 1) and rep is None:
            if is_date_before(ac_tmp_item[1], datetime.now()):
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('Start time is in the past')
                return 6

        if the_type in (0, 2) and rep is None:
            if is_date_before(ac_tmp_item[2], datetime.now()):
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('Stop time is in the past')
                return 7

        if is_date_before(ac_tmp_item[2], ac_tmp_item[1]):
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Start time < end time')
            return 8

        self._entry = tmp_item
        logger.error(f'{str(self._entry) = }')
        return 0

    def _thread_show_date(self, win, lock, go_on, stop):
        ''' display current date

            Parameters
            win     function to return the window to print to
            lock    threading lock
            go_on   function to indicate if we can print string
            stop    function to make the thread stop
        '''
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('_thread_show_date started!')
        d = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']
        while True:
            if go_on()():
                now = datetime.now()
                now_str = ' ' + now.strftime('%Y-%m-%d %H:%M:%S') + ', a ' + d[now.weekday()] + ' '
                X = int((self._maxX - len(now_str)) / 2) - 4
                if stop():
                    break
                w = win()
                if w:
                    with lock:
                        try:
                            self._win.addstr(5, 2, '─' * (self._maxX - 4), curses.color_pair(11))
                        except:
                            self._win.addstr(5, 2, '─'.encode('utf-8') * (self._maxX - 4), curses.color_pair(12))
                        if stop():
                            break
                        w.addstr(5, X, ' Now is:', curses.color_pair(10))
                        w.addstr(now_str, curses.color_pair(3))
                        w.refresh()
                for _n in range(5):
                    if stop():
                        break
                    sleep(.1)
            else:
                if stop():
                    break
                sleep(.1)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('_thread_show_date terminated!')

    def _set_start_session_date_time(self, a_date=None, time_delta=0):
        if a_date is None:
            time_delta = 0
            a_date = datetime.now()

        displ = 0 if self._focus < 8 else 6
        enable = range(2 + displ, 7 + displ)
        disable = (7 + displ, )
        checked = (2 + displ, 4 + displ)
        notchecked = (6 + displ, )
        f_date = 3 + displ
        f_time = f_date + 2
        f_dur = f_time + 2

        for n in enable:
            logger.error(f'setting {n} enabled to {True}')
            self._widgets[n].enabled = True
        for n in disable:
            logger.error(f'setting {n} enabled to {False}')
            self._widgets[n].enabled = False
        for n in checked:
            logger.error(f'setting {n} checked to {True}')
            self._widgets[n].checked = True
        for n in notchecked:
            logger.error(f'setting {n} checked to {False}')
            self._widgets[n].checked = False

        self._widgets[f_date].set_date_tupple((
                    a_date.year,
                    a_date.month,
                    a_date.day
                )
        )

        self._widgets[f_time].set_time_pyradio_time((
                a_date.hour,
                a_date.minute,
                a_date.second,
                0
            )
        )

        self._widgets[f_dur].set_time_pyradio_time((time_delta, 0, 0, 0))

    def keypress(self, char):
        logger.error(f'{self._widgets[18].enabled = }')
        '''
        PyRadioSimpleScheduleWindow keypress
        Return
        ======
            -1: Cancel
             0: Continue
             1: Get result
             2: Remove Schedule
             2: Show Help
             3: Invalid entry
             4: Open playlist selection window
             5: Open station selection window
             6: Start time is in the past
             7: Stop time is in the past
             8: Start time < end time
        '''
        l_char = None
        # if char == ord('0'):
        #     self._error_num = self._validate_selection()
        #     logger.error('self._error_num = {}'.format(self._error_num))
        #     return self._error_num

        # if char == ord('1'):
        #     self._error_num = self._validate_selection()
        #     logger.error('self._error_num = {}'.format(self._error_num))
        #     return self._error_num

        # logger.error(f'{self._focus = }')
        # logger.error(f'{char = }')

        l_char = None
        if self._widgets[self._focus].w_id == 22:
            ret = self._widgets[22].keypress(self._win, char)
            logger.error(f'return = {ret}')
            if ret == -1:
                self._stop = True
                self._exit = True
                return -1
            if ret == 2:
                # line editor help
                return 9

        elif char in self._global_functions or \
                (l_char := check_localized(char, self._global_functions.keys(), True)) is not None:
            if l_char is None:
                l_char = char
            self._global_functions[l_char]()
            return 0

        elif char == ord('n') and self._focus in range(2,14):
            # now
            self._set_start_session_date_time()
            if self._focus < 8:
                self._focus = 5
            else:
                self._focus = 11

        elif char in range(48, 58) and self._focus in range(2,14):
            now = datetime.now()
            time_delta = char - 48
            displ = 0 if self._focus < 8 else 6
            if self._focus < 8:
                now = now + timedelta(hours=time_delta)
                self._set_start_session_date_time(now, time_delta)
            else:
                enable = (8, 10, 12, 13)
                disable = (9, 11 )
                checked = (8, 12)
                notchecked = (10, )
                f_date = 9
                f_time = 11
                f_dur = 13
                self._focus = 12
                if self._widgets[4].checked:
                    t_date = list(self._widgets[3].date_tuple)
                    t_time = list(self._widgets[5].active_time)
                    if t_time[-1] == 2:
                        t_time[0] += 12
                    now = datetime(
                            year=t_date[0],
                            month=t_date[1],
                            day=t_date[2],
                            hour=t_time[0],
                            minute=t_time[1],
                            second=t_time[2]
                    )
                    now = now + timedelta(hours=time_delta)

                for n in enable:
                    self._widgets[n].enabled = True
                for n in disable:
                    self._widgets[n].enabled = False
                for n in checked:
                    logger.error(f'setting {n} checked to {True}')
                    self._widgets[n].checked = True
                for n in notchecked:
                    logger.error(f'setting {n} checked to {False}')
                    self._widgets[n].checked = False

                self._widgets[f_date].set_date_tupple((
                            now.year,
                            now.month,
                            now.day
                        )
                )

                self._widgets[f_time].set_time_pyradio_time((
                        now.hour,
                        now.minute,
                        now.second,
                        0
                    )
                )

                self._widgets[f_dur].set_time_pyradio_time((time_delta, 0, 0, 0))

        elif char in (ord('t'), ord('f')) and self._focus in (3, 5, 7, 9, 11, 13):
            # make date / time equal
            if self._focus == 3:
                if char == ord('t'):
                    self._widgets[9].date = str(self._widgets[3])
                else:
                    self._widgets[3].date = str(self._widgets[9])
            elif self._focus == 9:
                if char == ord('f'):
                    self._widgets[9].date = str(self._widgets[3])
                else:
                    self._widgets[3].date = str(self._widgets[9])
            elif self._focus == 5:
                if char == ord('t'):
                    self._widgets[11].set_time_pyradio_time(self._widgets[5].active_time)
                else:
                    self._widgets[5].set_time_pyradio_time(self._widgets[11].active_time)
            elif self._focus == 11:
                if char == ord('f'):
                    self._widgets[11].set_time_pyradio_time(self._widgets[5].active_time)
                else:
                    self._widgets[5].set_time_pyradio_time(self._widgets[11].active_time)
            elif self._focus == 7:
                if char == ord('t'):
                    self._widgets[13].set_time_pyradio_time(self._widgets[7].active_time)
                else:
                    self._widgets[7].set_time_pyradio_time(self._widgets[13].active_time)
            elif self._focus == 13:
                if char == ord('f'):
                    self._widgets[13].set_time_pyradio_time(self._widgets[7].active_time)
                else:
                    self._widgets[7].set_time_pyradio_time(self._widgets[13].active_time)

        elif char == kbkey['?'] or \
                check_localized(char, (kbkey['?'], )):
            return 2

        elif char in (curses.KEY_EXIT, kbkey['q'], 27) or \
                check_localized(char, (kbkey['q'], )):
            self._stop = True
            self._exit = True
            return -1

        elif self._focus == 0 and (char in (
                curses.KEY_ENTER, ord('\n'), ord('\r'),
                curses.KEY_RIGHT, kbkey['l'], kbkey['pause']) or \
                check_localized(char, (kbkey['pause'], kbkey['l']))) and \
                self._widgets[2].checked:
            return 4

        elif self._focus == 1 and (char in (
                curses.KEY_ENTER, ord('\n'), ord('\r'),
                curses.KEY_RIGHT, kbkey['l'], kbkey['pause'] or \
                check_localized(char, (kbkey['pause'], kbkey['l'])))
                ) and self._widgets[2].checked:
            return 5

        elif self._focus == 19 and (char in (
                kbkey['l'], kbkey['h'], curses.KEY_LEFT, curses.KEY_RIGHT) or \
                check_localized(char, (kbkey['l'], kbkey['h']))):
            if char in (kbkey['h'], curses.KEY_LEFT) or \
                check_localized(char, (kbkey['h'], )):
                self._repeat_index -= 1
                if self._repeat_index < 0 :
                    self._repeat_index = len(self._repeat) - 1
            else:
                self._repeat_index += 1
                if self._repeat_index == len(self._repeat):
                    self._repeat_index = 0
            self._widgets[19].string = self._repeat[self._repeat_index].ljust(self._max_repeat_len)

        elif char in (kbkey['k'], curses.KEY_UP) or \
                check_localized(char, (kbkey['k'], )):
            self._go_up()

        elif char in (kbkey['j'], curses.KEY_DOWN) or \
                check_localized(char, (kbkey['j'], )):
            self._go_down()

        elif char in (ord('\t'), 9, kbkey['tab']) or \
                check_localized(char, (kbkey['tab'], )):
            if self._widgets[self._focus].w_id in (3, 5, 7, 9, 11, 13):
                self._widgets[self._focus].keypress(char)
            else:
                self._next_widget()

        elif char in (curses.KEY_BTAB, kbkey['stab']) or \
                check_localized(char, (kbkey['stab'], )):
            if self._widgets[self._focus].w_id in (3, 5, 7, 9, 11, 13):
                self._widgets[self._focus].keypress(char)
            else:
                self._previous_widget()

        elif char == kbkey['info'] or \
                check_localized(char, (kbkey['info'], )):
            self._show_info()
            return 10

        else:
            ret = self._widgets[self._focus].keypress(char)
            # logger.error('ret = {}'.format(ret))

            if self._widgets[self._focus].w_id == 21 \
                    and not ret:
                ''' Cancel '''
                return -1

            if self._widgets[self._focus].w_id == 20 \
                    and not ret:
                ''' OK '''
                self._error_num = self._validate_selection()
                if self._error_num == 0:
                    logger.error('return 1')
                    # TODO: create an entry
                    return 1
                logger.error(f'return {self._error_num}')
                return self._error_num

            if self._widgets[self._focus].w_id in (3, 5, 7, 9, 11, 13):
                ''' Time
                      -1: Cancel
                       0: Continue
                       1: Show help
                '''

            elif self._widgets[self._focus].w_id == 14:
                if char in (kbkey['l'], curses.KEY_RIGHT) or \
                        check_localized(char, (kbkey['l'], )):
                    self._current_player_id += 1
                    if self._current_player_id == len(self._supported_players):
                        self._current_player_id = 0
                    self._widgets[14].string = self._supported_players[self._current_player_id]
                elif char in (kbkey['h'], curses.KEY_LEFT) or \
                        check_localized(char, (kbkey['h'], )):
                    self._current_player_id -= 1
                    if self._current_player_id < 0:
                        self._current_player_id = len(self._supported_players) - 1
                    self._widgets[14].string = self._supported_players[self._current_player_id]

            elif self._focus in (2, 4, 6, 10, 12):
                ''' Check Boxes '''
                if self._widgets[self._focus].checked:
                    if self._focus == 2:
                        self._widgets[0].string = self._schedule_item.item['playlist']
                        self._widgets[1].string = self._schedule_item.item['station']
                        self._widgets[0].enabled = True
                        self._widgets[1].enabled = True
                    elif self._focus == 4:
                        self._widgets[6].checked = False
                    elif self._focus == 6:
                        self._widgets[4].checked = False
                    elif self._focus == 10:
                        self._widgets[12].checked = False
                    elif self._focus == 12:
                        self._widgets[10].checked = False
                else:
                    if self._focus == 2:
                        self._widgets[0].string = 'Any playlist'
                        self._widgets[1].string = 'Any station'
                        self._widgets[0].enabled = False
                        self._widgets[1].enabled = False
                    elif self._focus == 4:
                        self._widgets[6].checked = True
                    elif self._focus == 6:
                        self._widgets[4].checked = True
                    elif self._focus == 10:
                        self._widgets[12].checked = True
                    elif self._focus == 12:
                        self._widgets[10].checked = True
            # if self._widgets[2].checked:
            #     self._widgets[3].enabled = self._widgets[4].checked
            # else:
            #     self._widgets[3].enabled = False
            # if self._widgets[8].checked:
            #     self._widgets[9].enabled = self._widgets[10].checked
            # else:
            #     self._widgets[9].enabled = False

        self.show()
        return 0

    def _display_time(self, stop, win, Y, X, lock):
        showed = False
        while True:
            if showed:
                for _ in range(0, 5):
                    sleep(.1)
                    if stop():
                        # if logger.isEnabledFor(logging.DEBUG):
                        #     logger.debug('File watch thread stopped on: {}'.format(a_file))

                        break
            disp = []
            disp.append((' ─────── ', curses.color_pair(3)))
            disp.append(('Current date: ' + format_date_to_iso8851(), curses.color_pair(11)))
            disp.append((' ────────', curses.color_pair(3)))
            if stop():
                break
            showed = True
            with lock:
                for i, n in enumerate(disp):
                    if i == 0:
                        try:
                            win().addstr(Y(), X(), n[0], n[1])
                        except:
                            win().addstr(Y(), X(), n[0].encode('utf-8'), n[1])
                    else:
                        try:
                            win().addstr(n[0], n[1])
                        except:
                            win().addstr(n[0].encode('utf-8'), n[1])
                win().refresh()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Time display thread exited!')

if __name__ == '__main__':

    pass
    # x = PyRadioTime.pyradio_time_diference_in_seconds(
    #     [1, 10, 15], [1, 12, 20]
    # )
    # print(x)
    # print(PyRadioTime.seconds_to_sting(x))
    # exit()

    # old = datetime.now()
    # now = old + timedelta(seconds=1*60 + 10)

    # import time
    # import os
    # while True:
    #     d = now - old
    #     old = datetime.now()
    #     os.system('clear')
    #     print('Remaining:', PyRadioTime.delta_to_sting(d))
    #     if old >= now:
    #         break
    #     time.sleep(.5)




# -*- coding: utf-8 -*-
#
# Windows SMTC backend using PyWinRT (winrt-*) + MediaPlayer.
#
import ctypes
import threading
import queue
import time
import traceback
import logging
import asyncio
import os
import urllib.parse

try:
    import ctypes.wintypes as wintypes
except Exception:
    wintypes = None

from .os_media_base import MediaControls

logger = logging.getLogger(__name__)

# --- COM/WinRT bootstrap (STA thread) ---

COINIT_APARTMENTTHREADED = 0x2

try:
    ole32 = ctypes.windll.ole32
    combase = ctypes.windll.combase
    user32 = ctypes.windll.user32
except Exception:
    ole32 = None
    combase = None
    user32 = None


def _log_error(msg):
    try:
        tb = traceback.format_exc()
        if tb and "NoneType: None" not in tb:
            logger.error("%s\n%s", msg, tb)
        else:
            logger.error("%s", msg)
    except Exception:
        pass


def _coinit_sta():
    try:
        if not ole32:
            return -1
        hr = ole32.CoInitializeEx(None, COINIT_APARTMENTTHREADED)
        return int(hr)
    except Exception:
        _log_error("OS-MEDIA: SMTC: CoInitializeEx failed")
        return -1


def _couninit():
    try:
        if ole32:
            ole32.CoUninitialize()
    except Exception:
        _log_error("OS-MEDIA: SMTC: CoUninitialize failed")


def _roinit():
    # RO_INIT_SINGLETHREADED = 0
    try:
        if not combase:
            return -1
        combase.RoInitialize.argtypes = [ctypes.c_int]
        combase.RoInitialize.restype = ctypes.c_long
        hr = combase.RoInitialize(0)
        return int(hr)
    except Exception:
        _log_error("OS-MEDIA: SMTC: RoInitialize failed")
        return -1


def _rouninit():
    try:
        if combase:
            combase.RoUninitialize()
    except Exception:
        _log_error("OS-MEDIA: SMTC: RoUninitialize failed")


def _pump_win_messages_once():
    # Minimal message pumping on STA.
    # Some WinRT event delivery can depend on it in desktop apps.
    try:
        if not user32 or not wintypes:
            return
        msg = wintypes.MSG()
        PM_REMOVE = 0x0001
        while user32.PeekMessageW(ctypes.byref(msg), None, 0, 0, PM_REMOVE):
            user32.TranslateMessage(ctypes.byref(msg))
            user32.DispatchMessageW(ctypes.byref(msg))
    except Exception:
        _log_error("OS-MEDIA: SMTC: message pump failed")


# --------------------------
# STA thread runner (WinRT)
# --------------------------

class _SMTCWinRTThread(object):
    def __init__(self, cmd_queue, state_queue):
        self.cmdq = cmd_queue
        self.stateq = state_queue

        self._thread = None
        self._stop = threading.Event()
        self._ready = threading.Event()

        # WinRT objects (STA thread only)
        self._player = None
        self._smtc = None
        self._updater = None
        self._media = None
        self._storage = None
        self._streams = None

        # Event token best-effort
        self._button_token = None

        # Ignore OS commands while connecting
        self._ignore_commands = True

        # Cached state to reduce churn
        self._last_title = None
        self._last_artist = None
        self._last_status = 'Stopped'
        self._last_can_next = True
        self._last_can_prev = True

        # Cached thumbnail path (absolute local file path)
        self._last_thumb = None

        # Async loop for WinRT async APIs (STA thread only)
        self._loop = None

    def start(self):
        try:
            if self._thread and self._thread.is_alive():
                return True
            self._stop.clear()
            self._ready.clear()
            self._thread = threading.Thread(target=self._run, name="pyradio-smtc-winrt", daemon=False)
            self._thread.start()
            self._ready.wait(5.0)
            return self._ready.is_set()
        except Exception:
            _log_error("OS-MEDIA: SMTC: thread start failed")
            return False

    def stop(self):
        try:
            self._stop.set()
            if self._thread and self._thread.is_alive():
                self._thread.join()
            self._thread = None
        except Exception:
            _log_error("OS-MEDIA: SMTC: thread stop failed")

    def set_ignore_commands(self, value):
        try:
            self._ignore_commands = True if value else False
        except Exception:
            _log_error("OS-MEDIA: SMTC: set_ignore_commands failed")

    def _run(self):
        try:
            _coinit_sta()
            _roinit()

            # Import WinRT namespaces on STA thread
            try:
                import winrt.windows.media as media
                import winrt.windows.storage as storage
                import winrt.windows.storage.streams as streams
                from winrt.windows.media.playback import MediaPlayer
                self._media = media
                self._storage = storage
                self._streams = streams

                # Create an event loop for WinRT async operations on this STA thread.
                try:
                    self._loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(self._loop)
                except Exception:
                    _log_error("OS-MEDIA: SMTC: asyncio loop init failed")
                    self._loop = None
            except Exception:
                _log_error("OS-MEDIA: SMTC: import winrt namespaces failed")
                self._ready.set()
                return

            # Create MediaPlayer and SMTC
            try:
                self._player = MediaPlayer()
                self._smtc = self._player.system_media_transport_controls
                self._updater = self._smtc.display_updater

                # Enable SMTC + base buttons
                self._smtc.is_enabled = True
                self._smtc.is_play_enabled = True
                self._smtc.is_pause_enabled = True
                self._smtc.is_stop_enabled = True

                # Nav caps (default false; updated by stateq)
                self._smtc.is_next_enabled = False
                self._smtc.is_previous_enabled = False

                # Hook buttons -> cmd_queue
                self._try_hook_buttons()

            except Exception:
                _log_error("OS-MEDIA: SMTC: MediaPlayer/SMTC init failed")

            # Signal ready even if partial
            self._ready.set()

            # STA loop: consume state updates
            while not self._stop.is_set():
                did_work = False

                # Coalesce updates: drain queue and keep last value per key.
                pending = {}

                for _ in range(256):
                    try:
                        changed = self.stateq.get_nowait()
                    except queue.Empty:
                        changed = None
                    if not changed:
                        break

                    did_work = True

                    # Merge dict updates; last write wins.
                    try:
                        for k in changed:
                            pending[k] = changed.get(k)
                    except Exception:
                        # If changed is not a dict for any reason, ignore it.
                        pass

                if pending:
                    try:
                        self._apply_changed(pending)
                    except Exception:
                        _log_error("OS-MEDIA: SMTC: apply_changed failed")

                _pump_win_messages_once()
                if not did_work:
                    time.sleep(0.01)


        except Exception:
            _log_error("OS-MEDIA: SMTC: STA thread crashed")
            try:
                self._ready.set()
            except Exception:
                pass
        finally:
            try:
                self._cleanup()
            except Exception:
                _log_error("OS-MEDIA: SMTC: cleanup failed")
            try:
                _rouninit()
            except Exception:
                _log_error("OS-MEDIA: SMTC: RoUninitialize failed")
            try:
                _couninit()
            except Exception:
                _log_error("OS-MEDIA: SMTC: CoUninitialize failed")

    def _try_hook_buttons(self):
        try:
            if not self._smtc or not self._media:
                return False
            media = self._media

            def _on_button(sender, args):
                try:
                    if self._ignore_commands:
                        return

                    btn = args.button
                    cmd = None

                    if btn == media.SystemMediaTransportControlsButton.PLAY:
                        cmd = "PLAY"
                    elif btn == media.SystemMediaTransportControlsButton.PAUSE:
                        cmd = "STOP"
                        # cmd = "PAUSE"
                    elif btn == media.SystemMediaTransportControlsButton.STOP:
                        cmd = "STOP"
                    elif btn == media.SystemMediaTransportControlsButton.NEXT:
                        cmd = "NEXT"
                    elif btn == media.SystemMediaTransportControlsButton.PREVIOUS:
                        cmd = "PREV"
                    elif btn == media.SystemMediaTransportControlsButton.PLAY_PAUSE:
                        cmd = "PLAYPAUSE"

                    if not cmd:
                        return

                    try:
                        self.cmdq.put_nowait((cmd, None))
                    except Exception:
                        # Never block STA thread
                        pass

                except Exception:
                    _log_error("OS-MEDIA: SMTC: button handler failed")

            # Prefer token-based add/remove
            tok = None
            try:
                tok = self._smtc.add_button_pressed(_on_button)
            except Exception:
                tok = None

            if tok is not None:
                self._button_token = tok
                return True

            # Fallback pattern if supported by projection
            try:
                self._smtc.button_pressed += _on_button
                self._button_token = None
                return True
            except Exception:
                _log_error("OS-MEDIA: SMTC: ButtonPressed hook failed")
                return False

        except Exception:
            _log_error("OS-MEDIA: SMTC: try_hook_buttons failed")
            return False


    def _normalize_thumb_path(self, value):
        # Accept absolute local path or file:// URL.
        # Returns "" if value is unusable.
        try:
            if not value:
                return ""
            v = str(value)
            if v.startswith("file://"):
                # file:///C:/... or file://C:/...
                v = v[7:]
                if v.startswith("/"):
                    # file:///C:/ -> /C:/ ; strip the leading slash
                    if len(v) > 2 and v[2] == ":":
                        v = v[1:]
                v = urllib.parse.unquote(v)
            v = v.strip().strip('"').strip("'")
            if not v:
                return ""
            # Only support absolute paths.
            if not os.path.isabs(v):
                return ""
            return v
        except Exception:
            _log_error("OS-MEDIA: SMTC: normalize_thumb_path failed")
            return ""

    def _load_thumb_stream_ref(self, abs_path):
        # Convert absolute file path to RandomAccessStreamReference (STA thread only).
        # Returns None on failure.
        _log_error("OS-MEDIA: SMTC: thumb start")
        try:
            if not abs_path:
                _log_error("OS-MEDIA: SMTC: Error 1")
                return None
            if not self._storage or not self._streams:
                _log_error("OS-MEDIA: SMTC: Error 2")
                return None
            if not self._loop:
                _log_error("OS-MEDIA: SMTC: Error 3")
                return None
            if not os.path.exists(abs_path):
                _log_error("OS-MEDIA: SMTC: Error 4")
                return None
            _log_error("OS-MEDIA: SMTC: thumb before async")
            async def _get_ref():
                f = await self._storage.StorageFile.get_file_from_path_async(abs_path)
                return self._streams.RandomAccessStreamReference.create_from_file(f)
            _log_error("OS-MEDIA: SMTC: Sending thumb")
            try:
                return self._loop.run_until_complete(_get_ref())
            except Exception:
                _log_error("OS-MEDIA: SMTC: load thumbnail stream failed")
                return None

        except Exception:
            _log_error("OS-MEDIA: SMTC: load_thumb_stream_ref failed")
            return None

    def _apply_changed(self, changed):
        # Expected keys based on MediaControls:
        #   PlaybackStatus: "Playing"/"Stopped"
        #   CanGoNext: bool
        #   CanGoPrevious: bool
        #   Volume: float 0..1   (SMTC does not own system volume; ignored)
        #   Metadata: dict (MPRIS-like or simplified)
        try:
            if not self._smtc:
                return

            media = self._media

            if "CanGoNext" in changed:
                v = True if changed.get("CanGoNext") else False
                logger.debug('OS-MEDIA: SMTC: \n\n')
                logger.debug(f'OS-MEDIA: SMTC: {self._last_can_next = } -/- {v = }')
                if v != self._last_can_next:
                    self._last_can_next = v
                    try:
                        self._smtc.is_next_enabled = v
                        logger.debug('OS-MEDIA: SMTC: CanGoNext updated\n\n')
                    except Exception:
                        _log_error("OS-MEDIA: SMTC: set is_next_enabled failed")

            if "CanGoPrevious" in changed:
                v = True if changed.get("CanGoPrevious") else False
                logger.debug('OS-MEDIA: SMTC: \n\n')
                logger.debug(f'OS-MEDIA: SMTC: {self._last_can_prev = } -/- {v = }')
                if v != self._last_can_prev:
                    self._last_can_prev = v
                    try:
                        self._smtc.is_previous_enabled = v
                        logger.debug('OS-MEDIA: SMTC: CanGoPrevious updated\n\n')
                    except Exception:
                        _log_error("OS-MEDIA: SMTC: set is_previous_enabled failed")

            if "PlaybackStatus" in changed:
                logger.debug('OS-MEDIA: SMTC: \n\n')
                st = str(changed.get("PlaybackStatus") or "")
                logger.debug(f'OS-MEDIA: SMTC: {self._last_status = } -/- {st= }')
                if st != self._last_status:
                    self._last_status = st
                    try:
                        if st == "Playing":
                            self._smtc.playback_status = media.MediaPlaybackStatus.PLAYING
                            logger.debug('OS-MEDIA: SMTC: PlaybackStatus PLAYING updated\n\n')
                        elif st == "Paused":
                            self._smtc.playback_status = media.MediaPlaybackStatus.PAUSED
                            logger.debug('OS-MEDIA: SMTC: PlaybackStatus PAUSED updated\n\n')
                        else:
                            self._smtc.playback_status = media.MediaPlaybackStatus.STOPPED
                            logger.debug('OS-MEDIA: SMTC: PlaybackStatus STOPPED updated\n\n')
                    except Exception:
                        _log_error("OS-MEDIA: SMTC: set playback_status failed")

            if "Metadata" in changed:
                md = changed.get("Metadata") or {}

                # Accept both:
                # - simplified: {"title": "...", "artist": "..."}
                # - mpris-like: {"xesam:title": "...", "xesam:artist": ["..."], ...}
                title = ""
                artist = ""

                try:
                    if "title" in md or "artist" in md:
                        title = md.get("title") or ""
                        artist = md.get("artist") or ""
                        playlist = md.get("album") or ""
                    else:
                        title = md.get("xesam:title") or ""
                        a = md.get("xesam:artist")
                        playlist = md.get("xesam:album") or ""
                        if isinstance(a, list) and a:
                            artist = str(a[0] or "")
                        else:
                            artist = ""
                except Exception:
                    title = ""
                    artist = ""
                # Thumbnail support (local absolute file path).
                # Accept md['thumbnail'] or md['mpris:artUrl'] (file path or file:// URL).
                thumb = ""
                try:
                    if "thumbnail" in md:
                        thumb = md.get("thumbnail") or ""
                    elif "mpris:artUrl" in md:
                        thumb = md.get("mpris:artUrl") or ""
                except Exception:
                    thumb = ""

                thumb_path = self._normalize_thumb_path(thumb)
                logger.debug(f'OS-MEDIA: SMTC: {thumb_path = }')

                logger.debug('OS-MEDIA: SMTC: \n\n')
                logger.debug(f'OS-MEDIA: SMTC: {title = } -/- {self._last_title = }')
                logger.debug(f'OS-MEDIA: SMTC: {artist = } -/- {self._last_artist = }')
                meta_changed = ((title != self._last_title) or (artist != self._last_artist) or (thumb_path != (self._last_thumb or "")))
                if meta_changed:
                    self._last_title = title
                    self._last_artist = artist
                    try:
                        if self._updater:
                            self._updater.type = media.MediaPlaybackType.MUSIC
                            props = self._updater.music_properties
                            props.title = title
                            props.artist = artist
                            props.album_title = playlist

                            # Apply thumbnail only when it changes (expensive async work).
                            try:
                                if thumb_path != (self._last_thumb or ""):
                                    self._last_thumb = thumb_path
                                    _log_error(f"OS-MEDIA: SMTC: thumbnail final {thumb_path = }")
                                    if not thumb_path:
                                        self._updater.thumbnail = None
                                        _log_error("OS-MEDIA: SMTC: thumbnail is None")
                                    else:
                                        ref = self._load_thumb_stream_ref(thumb_path)
                                        _log_error(f"OS-MEDIA: SMTC: thumbnail {ref = }")
                                        if ref is not None:
                                            self._updater.thumbnail = ref
                                            _log_error(f"OS-MEDIA: SMTC: thumbnail updated")
                            except Exception:
                                _log_error("OS-MEDIA: SMTC: set thumbnail failed")
                            self._updater.update()
                            logger.debug('OS-MEDIA: SMTC: update sent\n\n')
                    except Exception:
                        _log_error("OS-MEDIA: SMTC: DisplayUpdater update failed")
                else:
                    logger.debug('OS-MEDIA: SMTC: no update sent\n\n')
            # Volume is ignored on purpose.

        except Exception:
            _log_error("OS-MEDIA: SMTC: _apply_changed failed")

    def _cleanup(self):
        # Deterministic shutdown best-effort.
        # This runs on the STA thread.
        try:
            # Disable SMTC to reduce stale sessions after shutdown.
            # Even if this produces a final change event, it avoids lingering controls.
            try:
                if self._smtc:
                    self._smtc.is_enabled = False
                    self._smtc.is_play_enabled = False
                    self._smtc.is_pause_enabled = False
                    self._smtc.is_stop_enabled = False
                    self._smtc.is_next_enabled = False
                    self._smtc.is_previous_enabled = False
            except Exception:
                _log_error("OS-MEDIA: SMTC: disable SMTC failed")

            # Unsubscribe from ButtonPressed if we have a token
            try:
                if self._smtc and self._button_token is not None:
                    try:
                        self._smtc.remove_button_pressed(self._button_token)
                    except Exception:
                        _log_error("OS-MEDIA: SMTC: remove_button_pressed failed")
            except Exception:
                _log_error("OS-MEDIA: SMTC: button cleanup failed")

            self._button_token = None

            # Release WinRT objects (best-effort)
            self._updater = None
            self._smtc = None
            self._player = None
            self._media = None
            self._storage = None
            self._streams = None

            # Close asyncio loop (if any)
            try:
                if self._loop is not None:
                    try:
                        self._loop.stop()
                    except Exception:
                        pass
                    try:
                        self._loop.close()
                    except Exception:
                        pass
            except Exception:
                _log_error("OS-MEDIA: SMTC: asyncio loop cleanup failed")
            self._loop = None

        except Exception:
            _log_error("OS-MEDIA: SMTC: cleanup failed")
# --------------------------
# Public controller (owned by main thread)
# --------------------------

class WindowsSMTCController(MediaControls):
    """
    Main-thread owner, mirroring MprisController semantics.
    """

    def __init__(self, identity="PyRadio", instance_name=None, default_icon=None):
        super().__init__(identity, instance_name, default_icon)
        self._thread = _SMTCWinRTThread(self._cmdq, self._stateq)

    def start(self):
        try:
            ok = self._thread.start()
            if not ok:
                logger.error("OS-MEDIA: SMTC: start failed")
                return False

            # Ignore OS commands while we publish the first snapshot
            self._thread.set_ignore_commands(True)

            # Best-effort initial snapshot (similar to how MPRIS starts with defaults)
            try:
                self._stateq.put({"PlaybackStatus": "Stopped"})
                self._stateq.put({"CanGoPrevious": False, "CanGoNext": False})
                # Provide an empty Metadata block so updater is initialized
                self._stateq.put({"Metadata": {"xesam:title": "", "xesam:artist": []}})
            except Exception:
                _log_error("OS-MEDIA: SMTC: initial snapshot failed")

            self._thread.set_ignore_commands(False)
            return True

        except Exception:
            _log_error("OS-MEDIA: SMTC: controller start failed")
            return False

    def stop(self):
        try:
            self._thread.set_ignore_commands(True)
            self._thread.stop()
        except Exception:
            _log_error("OS-MEDIA: SMTC: controller stop failed")

    # ------------- main-loop servicing -------------

    def poll(self, enabled):
        """
        Same behavior as MprisController.poll(enabled):
        - if disabled: drain all pending commands and log once
        - if enabled: drain all commands quickly and execute callbacks
        - handle debounced SET_VOLUME via base class
        """
        if not enabled:
            had_cmd = False
            while True:
                try:
                    self._cmdq.get_nowait()
                    had_cmd = True
                except queue.Empty:
                    break
            if had_cmd and logger.isEnabledFor(logging.INFO):
                logger.info("**** SMTC actions are allowed in NORMAL MODE only")
            super().poll(False)
            return

        while True:
            try:
                cmd, arg = self._cmdq.get_nowait()
            except queue.Empty:
                break

            if cmd == "PLAY" and self.cb_play:
                self.cb_play()

            elif cmd in ("STOP", "PAUSE") and self.cb_stop:
                self.cb_stop()

            elif cmd == "NEXT" and self.cb_next:
                self.cb_next()

            elif cmd == "PREV" and self.cb_prev:
                self.cb_prev()

            elif cmd == "PLAYPAUSE":
                if self.cb_playpause:
                    self.cb_playpause()
                else:
                    # Keep the same default policy as MPRIS: ignore.
                    pass

            elif cmd == "SET_VOLUME":
                # latest-wins (debounced) - identical to MPRIS
                try:
                    self._vol_pending = int(arg)
                except Exception:
                    self._vol_pending = None
                self._vol_last_req_ts = time.monotonic()

        super().poll(enabled)

    # ------------- state updates (main -> STA thread) -------------

    def update_metadata(self, trackid, title, station_name, playlist_name, url=None, art_url=None):
        """
        Keep the same metadata shape as MPRIS, but only title/artist are used by SMTC.
        """
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'OS_MEDIA: SMTC: {trackid = }')
            logger.debug(f'OS-MEDIA: SMTC: {title = }')
            logger.debug(f'OS-MEDIA: SMTC: {station_name = }')
            logger.debug(f'OS-MEDIA: SMTC: {playlist_name = }')
            logger.debug(f'OS-MEDIA: SMTC: {url = }')
            logger.debug(f'OS-MEDIA: SMTC: {art_url = }')
        if art_url is not None:
            art_url = art_url[7:].lstrip("/").replace("/", os.sep)
            logger.error(f'OS-MEDIA: SMTC: fixed {art_url = }')
        try:
            md = {}
            md["mpris:trackid"] = trackid
            md["xesam:title"] = title
            md["xesam:artist"] = [station_name] if station_name else []
            md["xesam:album"] = playlist_name if playlist_name else ""
            if url:
                md["xesam:url"] = url
            if art_url:
                md["mpris:artUrl"] = art_url

            self._stateq.put({"Metadata": md})
        except Exception:
            _log_error("OS-MEDIA: SMTC: update_metadata failed")





# -*- coding: utf-8 -*-
"""
Playlist validation utilities with threading and host-aware throttling
"""
import sys
import locale
import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from urllib.parse import urlparse
from collections import defaultdict
import requests

from .common import Station, RichColorPrinter

locale.setlocale(locale.LC_ALL, "")


def check_url(url, referer_url, semaphore, timeout=5, read_bytes=1024):
    """Check if URL provides actual playable audio stream using requests"""
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36',
        'Accept': 'audio/*, video/*, */*',
        'Accept-Language': 'en-US,en;q=0.5',
        'Icy-MetaData': '1'
    }
    if referer_url:
        headers['Referer'] = referer_url

    try:
        with semaphore:
            # USE REQUESTS INSTEAD OF URLLIB
            response = requests.get(
                url,
                headers=headers,
                timeout=timeout,
                stream=True,  # Important for audio streams
                allow_redirects=True  # AUTO-REDIRECTS!
            )
            response.raise_for_status()  # Raise exception for bad status codes

            # 1. READ FIRST BYTES
            data = b''
            for chunk in response.iter_content(chunk_size=read_bytes):
                if chunk:
                    data = chunk
                    break

            # 2. CHECK FOR HTML RESPONSES
            is_html = (data.startswith(b'<html') or
                      b'<!DOCTYPE' in data or
                      b'<xml' in data or
                      b'<!' in data[:100])

            if is_html:
                return False  # Server returned HTML, not audio

            # 3. CHECK CONTENT-TYPE - MUST be audio/video
            content_type = response.headers.get('Content-Type', '').lower()
            is_media = any(media_type in content_type for media_type in
                         ['audio/', 'video/', 'application/ogg', 'application/vnd.apple.mpegurl'])

            if is_media:
                return True  # Content-Type confirms it's media

            # 4. CHECK FOR ICY METADATA (Shoutcast/Icecast)
            icy_headers = any(response.headers.get(h) for h in
                            ['icy-name', 'icy-genre', 'icy-url'])

            if icy_headers:
                return True  # ICY headers confirm it's a stream

            # 5. FINAL CHECK: VERIFY AUDIO SIGNATURES
            return detect_audio_signatures(data)

    except requests.exceptions.RequestException:
        return False
    except Exception:
        return False

def detect_audio_signatures(data):
    """Check for actual audio stream signatures"""
    # Common audio format headers
    audio_patterns = [
        (b'ID3', 0),           # MP3 metadata (start of data)
        (b'\xFF\xFB', 0),      # MP3 frame sync
        (b'\xFF\xF1', 0),      # AAC frame sync
        (b'OggS', 0),          # OGG container
        (b'fLaC', 0),          # FLAC
        (b'#EXTM3U', 0),       # HLS playlist
        (b'RIFF', 0),          # WAV
        (b'\x00\x00\x00\x18ftyp', 4),  # MP4
    ]

    for pattern, offset in audio_patterns:
        if len(data) > offset + len(pattern) and data[offset:offset+len(pattern)] == pattern:
            return True

    # For streams that might not have clear headers, check for non-text data
    if len(data) > 100:
        text_chars = bytearray({7,8,9,10,12,13,27} | set(range(0x20, 0x100)) - {0x7f})
        is_likely_binary = bool(data.translate(None, text_chars))
        return is_likely_binary

    return False

def print_statistics(total_count, online_count, groups_count, printer):
    """
    Print validation statistics with centered alignment
    """
    # Calculate statistics
    stations_count = total_count - groups_count
    offline_count = stations_count - online_count
    success_rate = (online_count / stations_count * 100) if stations_count > 0 else 0

    # Define all labels and values
    stats_data = [
        ("Online Stations:", online_count),
        ("Offline Stations:", offline_count),
        ("Groups:", groups_count),
        ("Total Stations:", stations_count),
        ("Total Items:", total_count)
    ]

    # Find maximum widths
    max_label_width = max(len(label) for label, _ in stats_data)
    max_number_width = max(len(str(value)) for _, value in stats_data)
    max_success_width = len(f"{success_rate:.1f}%")

    # Calculate total block width and padding
    total_block_width = max_label_width + 1 + max_number_width
    left_padding = (50 - total_block_width) // 2

    # Center title
    title = "VALIDATION SUMMARY"
    title_padding = (50 - len(title)) // 2

    printer("[cyan]" + "="*50 + "[/cyan]")
    printer(" " * title_padding + f"[bold]{title}[/bold]")
    printer("[cyan]" + "="*50 + "[/cyan]")

    # Print each statistic line
    for label, value in stats_data:
        padded_label = label.rjust(max_label_width)
        line = " " * left_padding + f"[green]{padded_label}[/green] [cyan]{value}[/cyan]"
        printer(line)

    # Center separator
    printer("[cyan]" + "-"*50 + "[/cyan]")

    # Center success rate
    success_label = "Success Rate:"
    padded_label = success_label.rjust(max_label_width)
    success_line = " " * left_padding + f"[bold]{padded_label}[/bold] [cyan]{success_rate:.1f}%[/cyan]"
    printer(success_line)

    printer("[cyan]" + "="*50 + "[/cyan]")

def check_playlist(file_path, mode="mark", threads=5, timeout=5, max_per_host=2, with_date=False, no_color=False, verbose=True):
    """Validate a CSV or M3U playlist with host-aware throttling"""

    items = []
    input_type = None

    # Create printer instance
    printer = RichColorPrinter(use_color=not no_color)

    if file_path.endswith(".csv"):
        from .common import CsvReadWrite
        reader = CsvReadWrite(file_path)
        if not reader.read():
            raise RuntimeError("Error reading CSV")
        items = reader.items
        input_type = "csv"

    elif file_path.endswith(".m3u"):
        from .m3u import parse_m3u
        items, error = parse_m3u(file_path)
        if error:
            printer(f'[red]Error:[/red] {error}')
            sys.exit(1)
        input_type = "m3u"

    else:
        raise ValueError("Unsupported file type")

    # Generate timestamp if requested
    timestamp = ""
    if with_date:
        from datetime import datetime
        timestamp = datetime.now().strftime("-%Y-%m-%d-%H-%M-%S")

    printer(f"[cyan]Checking {len(items)} stations using {threads} threads...[/cyan]")

    # --- host-aware throttling setup ---
    host_map = defaultdict(list)
    for idx, st in enumerate(items):
        host = urlparse(st[Station.url]).hostname
        host_map[host].append((idx, st))

    host_semaphores = {host: threading.BoundedSemaphore(max_per_host) for host in host_map}

    # --- threaded check ---
    results_dict = {}   # idx -> (station, ok_flag)
    futures = {}

    with ThreadPoolExecutor(max_workers=threads) as executor:
        for idx, st in enumerate(items):
            # Check if it's a group header
            if st[Station.url] == '-':  # Group header
                results_dict[idx] = (st, True)  # Use None for group headers
                if verbose:
                    printer(f"[cyan]{idx+1}[/cyan]. [cyan]GROUP[/cyan] {st[Station.name]}")
                continue

            # do not check group headers
            url = st[Station.url].replace('https://', 'http://') if st[Station.http] else st[Station.url]
            semaphore = host_semaphores[urlparse(url).hostname]
            futures[executor.submit(check_url, url, st[Station.referer], semaphore, timeout, 1024)] = (idx, st)

        for future in as_completed(futures):
            idx, st = futures[future]
            ok = False
            try:
                ok = future.result()
            except Exception:
                ok = False

            # store result in dict
            results_dict[idx] = (st, ok)

            # print immediately
            if verbose:
                if ok:
                    printer(f"[green]{idx+1}[/green]. [green]OK[/green] {st[Station.name]} ({st[Station.url]})")
                else:
                    printer(f"[red]{idx+1}[/red]. [red]BROKEN[/red] {st[Station.name]} ({st[Station.url]})")

    # --- collect results in original order ---
    ok_items, bad_items = [], []
    for idx in range(len(items)):
        st, ok_flag = results_dict[idx]
        if ok_flag:
            ok_items.append(st)
        else:
            bad_items.append(st)

    # --- drop mode ---
    if mode == "drop":
        ok_file = f"{file_path}.ok{timestamp}.{input_type}"
        bad_file = f"{file_path}.bad{timestamp}.{input_type}"
        if input_type == "csv":
            from .common import CsvReadWrite
            ret = CsvReadWrite(ok_file).write(items=ok_items)
            if ret < 0:
                printer(f'[red]Error:[/red] Cannot write "{ok_file}".')
                sys.exit(1)
            printer(f'[green]Saved working stations to "{ok_file}"[/green]')
            ret = CsvReadWrite(bad_file).write(items=bad_items)
            if ret < 0:
                printer(f'[red]Error:[/red] Cannot write "{bad_file}".')
                sys.exit(1)
            printer(f'[yellow]Saved failed stations to "{bad_file}"[/yellow]')
        else:
            from .m3u import list_to_m3u
            error = list_to_m3u(ok_items, ok_file)
            if error:
                printer(f'[red]Error:[/red] Cannot write "{ok_file}"\n:  [red]{error}[/red].')
                sys.exit(1)
            printer(f'[green]Saved working stations to "{ok_file}"[/green]')
            error = list_to_m3u(bad_items, bad_file)
            if error:
                printer(f'[red]Error:[/red] Cannot write "{bad_file}"\n:  [red]{error}[/red].')
                sys.exit(1)
            printer(f'[yellow]Saved failed stations to "{bad_file}"[/yellow]')

        print_statistics(
            len(items),
            len(ok_items),
            sum(1 for st in items if st[Station.url] == '-'),
            printer
        )

        return

    # --- mark mode (default) ---
    marked_items = []
    for idx, st in enumerate(items):
        _, ok_flag = results_dict[idx]
        if not ok_flag:
            st = list(st)  # make copy
            st[Station.name] = "[X] " + st[Station.name]
        marked_items.append(st)

    out_file = f"{file_path}.validated{timestamp}.{input_type}"
    if input_type == "csv":
        from .common import CsvReadWrite
        ret = CsvReadWrite(out_file).write(items=marked_items)
        if ret < 0:
            printer(f'[red]Error:[/red] Cannot write "{out_file}".')
            sys.exit(1)
    else:
        from .m3u import list_to_m3u
        error = list_to_m3u(marked_items, out_file)
        if error:
            printer(f'[red]Error:[/red] Cannot write "{out_file}".')
            sys.exit(1)

    group_count = sum(1 for st in items if st[Station.url] == '-')
    online_count = sum(1 for st in marked_items
                  if st[Station.url] != '-' and not st[Station.name].startswith("[X]"))
    print_statistics(
        len(items),
        online_count,
        group_count,
        printer
    )

    printer(f'[cyan]Validation results written to "{out_file}"[/cyan]')



# -*- coding: utf-8 -*-
import curses
import curses.ascii
import platform
from collections import OrderedDict
from collections.abc import Iterable
import json
import logging
import locale
import string
import curses
from os.path import join, exists, dirname
from collections import deque
from threading import Lock
import re
try:
    from importlib.resources import files, as_file   # 3.9+
except ImportError:
    from importlib_resources import files, as_file   # backport για 3.7–3.8
try:
    from .cjkwrap import is_wide
except ImportError:
    pass

locale.setlocale(locale.LC_ALL, '')    # set your locale

logger = logging.getLogger(__name__)

_KB2STR_PATTERN = re.compile(r'\{([^{}]+)\}')

input_queue = deque()
queue_lock = Lock()

def enqueue_input(char, reset=False):
    global input_queue
    if reset:
        input_queue.clear()
    with queue_lock:
        input_queue.append(char)

def dequeue_input():
    global input_queue
    with queue_lock:
        if input_queue:
            return input_queue.popleft()
        return None

def clear_input_queue():
    global input_queue
    with queue_lock:
        input_queue.clear()

kb_letter = ''
kb_cjk = False

kbkey_orig = OrderedDict()
# ! Movement Keys
kbkey_orig['h_movement']               = ( None                   , 'Movement Keys')
kbkey_orig['j']                        = ( ord('j')               , 'Go down')
kbkey_orig['k']                        = ( ord('k')               , 'Go up')
kbkey_orig['h']                        = ( ord('h')               , 'Go left')
kbkey_orig['l']                        = ( ord('l')               , 'Go right')
kbkey_orig['g']                        = ( ord('g')               , 'Go to top of list')
kbkey_orig['G']                        = ( ord('G')               , 'Go to end of list')
kbkey_orig['screen_top']               = ( ord('H')               , 'Go to top of screen')
kbkey_orig['screen_middle']            = ( ord('M')               , 'Go to middle of screen')
kbkey_orig['screen_bottom']            = ( ord('L')               , 'Go to bottom of screen')

# ! Volume Keys
kbkey_orig['h_volume']                 = ( None                   , 'Volume Keys')
kbkey_orig['v_up1']                    = ( ord('+')               , 'Volume up Key 1')                                # global
kbkey_orig['v_up2']                    = ( ord('.')               , 'Volume up Key 2')                                # global
kbkey_orig['v_up3']                    = ( ord('=')               , 'Volume up Key 3')                                # global
kbkey_orig['v_dn1']                    = ( ord(',')               , 'Volume down Key 1')                              # global
kbkey_orig['v_dn2']                    = ( ord('-')               , 'Volume down Key 2')                              # global
kbkey_orig['mute']                     = ( ord('m')               , 'Mute player')                                    # global
kbkey_orig['s_vol']                    = ( ord('v')               , 'Save volume')                                    # global

# ! Global / Multi Window Keys
kbkey_orig['h_global']                 = ( None                   , 'Global / Multi Window Keys')
kbkey_orig['?']                        = ( ord('?')               , 'Open help window')
kbkey_orig['s']                        = ( ord('s')               , 'Save, Accept, RadioBrowser search, etc.')
kbkey_orig['q']                        = ( ord('q')               , 'Exit or Cancel')
kbkey_orig['y']                        = ( ord('y')               , 'Answer Yes')
kbkey_orig['Y']                        = ( ord('Y')               , 'Answer Yes to All')
kbkey_orig['n']                        = ( ord('n')               , 'Answer No')
kbkey_orig['N']                        = ( ord('N')               , 'Answer No to All')
kbkey_orig['del']                      = ( ord('x')               , 'Delete an item')
kbkey_orig['paste']                    = ( ord('p')               , 'Paste')
kbkey_orig['t']                        = ( ord('t')               , 'Open themes window')
kbkey_orig['transp']                   = ( ord('T')               , 'Toggle transparency')                            # global
kbkey_orig['revert_saved']             = ( ord('r')               , 'Revert to saved values')
kbkey_orig['revert_def']               = ( ord('d')               , 'Revert to default values')
kbkey_orig['tab']                      = ( ord('L')               , 'Alternative Tab')
kbkey_orig['stab']                     = ( ord('H')               , 'Alternative Shift-Tab')
kbkey_orig['no_show']                  = ( ord('x')               , 'Do not show Info / Warning message again')
kbkey_orig['tag']                      = ( ord('w')               , 'Tag a title')                                    # global
kbkey_orig['t_tag']                    = ( ord('W')               , 'Toggle Titles Tagging')                          # global
kbkey_orig['next']                     = ( ord('n')               , 'Go to next item')
kbkey_orig['prev']                     = ( ord('p')               , 'Go to previous item')
kbkey_orig['no_buffer']                = ( ord('z')               , 'Buffering Window > Set to 0 (disable)')
kbkey_orig['repaint']                  = ( ord('#')               , 'Repaint screen')
kbkey_orig['info_rename']              = ( ord('r')               , 'Info Window > Rename station')
kbkey_orig['reload']                   = ( ord('r')               , 'Reload from disk')
kbkey_orig['watch_theme']              = ( ord('c')               , 'Themes Window > Watch theme for changes')

# ! Main Window keys
kbkey_orig['h_main']                   = ( None                   , 'Main Window keys')
kbkey_orig['open_config']              = ( ord('c')               , 'Open config window')
kbkey_orig['open_playlist']            = ( ord('o')               , 'Open playlists list')
kbkey_orig['open_online']              = ( ord('O')               , 'Open online services (Radio Browser)')
kbkey_orig['open_enc']                 = ( ord('E')               , 'Open encodings window')
kbkey_orig['extra_p_pamars']           = ( ord('Z')               , 'Open "Player Extra Parameters" window')
kbkey_orig['edit']                     = ( ord('e')               , 'Edit an item')
kbkey_orig['add']                      = ( ord('a')               , 'Add item (station, group, whatever)')
kbkey_orig['append']                   = ( ord('A')               , 'Append item (add to end of current list)')
kbkey_orig['gr']                       = ( curses.ascii.BEL       , 'Open groups window')                             # default: ^G
kbkey_orig['gr_next']                  = ( curses.ascii.EM        , 'Go to next group')                               # default: ^E
kbkey_orig['gr_prev']                  = ( curses.ascii.ENQ       , 'Go to previous group')                           # default: ^Y
kbkey_orig['open_regs']                = ( ord('\'')              , 'Open registers list')
kbkey_orig['open_extra']               = ( ord('\\')              , 'Open extra commands')
kbkey_orig['add_to_reg']               = ( ord('y')               , 'Add station to register')
kbkey_orig['info']                     = ( ord('i')               , 'Display station info')
kbkey_orig['fav']                      = ( ord('*')               , 'Add station to favorites')
kbkey_orig['https']                    = ( ord('z')               , 'Toggle force use https')
kbkey_orig['pause']                    = ( ord(' ')               , 'Stop playback (Pause if recording)')
kbkey_orig['random']                   = ( ord('r')               , 'Play random station')
kbkey_orig['p_next']                   = ( curses.ascii.SO        , 'Play next station')                              # default: ^N
kbkey_orig['p_prev']                   = ( curses.ascii.DLE       , 'Play next station')                              # default: ^P
kbkey_orig['Reload']                   = ( ord('R')               , 'Reload current playlist')
kbkey_orig['fuzzy_search']             = ( ord('f')               , 'Open Fuzzy Search')
kbkey_orig['t_calc_col']               = ( ord('~')               , 'Toggle calculated colors')
kbkey_orig['rec']                      = ( ord('|')               , 'Toggle recording')
kbkey_orig['jump']                     = ( ord('J')               , 'Create a Jump tag')
kbkey_orig['goto_playing']             = ( ord('P')               , 'Go to playing station')
kbkey_orig['st_up']                    = ( curses.ascii.NAK       , 'Move station up')                                # default: ^U
kbkey_orig['st_dn']                    = ( curses.ascii.EOT       , 'Move station down')                              # default: ^D
kbkey_orig['hist_next']                = ( ord('>')               , 'Play next history item')
kbkey_orig['hist_prev']                = ( ord('<')               , 'Play previous history item')
kbkey_orig['ext_player']               = ( ord('X')               , 'Launch External Player')

# ! Search function
kbkey_orig['h_search']                 = ( None                   , 'Search function')
kbkey_orig['search']                   = ( ord('/')               , 'Open search subwindow')
kbkey_orig['search_next']              = ( ord('n')               , 'Search down')
kbkey_orig['search_prev']              = ( ord('N')               , 'Search up')



# ! Extra Commands Keys:)
kbkey_orig['h_extra']                  = ( None                   , 'Extra Commands Keys')
kbkey_orig['new_playlist']             = ( ord('n')               , 'Create a new playlist')
kbkey_orig['rename_playlist']          = ( ord('r')               , 'Rename current playlist')
kbkey_orig['open_remote_control']      = ( ord('s')               , 'Open "PyRadio Remote Control" window')
kbkey_orig['open_dirs']                = ( ord('o')               , 'Open dirs in file manager')
kbkey_orig['change_player']            = ( ord('m')               , 'Cahnge media player')
kbkey_orig['hist_top']                 = ( ord(']')               , 'Open first opened playlist')
kbkey_orig['buffer']                   = ( ord('b')               , 'Toggle buffering')
kbkey_orig['toggle_time']              = ( ord('t')               , 'Toggle displaying time')
kbkey_orig['toggle_tts']               = ( ord('T')               , 'Toggle TTS')
kbkey_orig['open_buffer']              = ( ord('B')               , 'Open buffering window')
kbkey_orig['last_playlist']            = ( ord('l')               , 'Toggle Open last playlist')
kbkey_orig['clear_reg']                = ( ord('c')               , 'Clear current register')
kbkey_orig['clear_all_reg']            = ( ord('C')               , 'Clear all registers')
kbkey_orig['unnamed']                  = ( ord('u')               , 'Show unnamed register')
kbkey_orig['html_help']                = ( ord('h')               , 'Display html help')
kbkey_orig['station_volume']           = ( ord('v')               , 'Save station volume')
kbkey_orig['toggle_station_volume']    = ( ord('V')               , 'Toggle use of station volume')

# ! RadioBrowser Keys:
kbkey_orig['h_rb']                     = ( None                   , 'RadioBrowser Keys')
kbkey_orig['rb_vote']                  = ( ord('V')               , 'Vote for station')
kbkey_orig['rb_info']                  = ( ord('I')               , 'Station DB info')
kbkey_orig['rb_server']                = ( ord('C')               , 'Select server to connect to')
kbkey_orig['rb_sort']                  = ( ord('S')               , 'Sort search results')
kbkey_orig['rb_p_first']               = ( ord('{')               , 'Go to first search results page')
kbkey_orig['rb_p_next']                = ( ord(']')               , 'Go to next search results page')
kbkey_orig['rb_p_prev']                = ( ord('[')               , 'Go to previous search results page')

# ! RadioBrowser Search Keys
kbkey_orig['h_rb_s']                   = ( None                   , 'RadioBrowser Search Window Keys')
kbkey_orig['rb_h_next']                = ( curses.ascii.SO        , 'Go to next item')                     # default: ^N
kbkey_orig['rb_h_prev']                = ( curses.ascii.DLE       , 'Go to previous item')                 # default: ^P
kbkey_orig['rb_h_add']                 = ( curses.ascii.EM        , 'Add item')                            # default: ^Y
kbkey_orig['rb_h_del']                 = ( curses.ascii.CAN       , 'Delete item')                         # default: ^X
kbkey_orig['rb_h_def']                 = ( curses.ascii.STX       , 'Make item default')                   # default: ^B
kbkey_orig['rb_h_0']                   = ( curses.ascii.ACK       , 'Go to template (item 0)')             # default: ^F
kbkey_orig['rb_h_save']                = ( curses.ascii.ENQ       , 'Save items')                          # default: ^E

# ! Keyboard Window Keys
kbkey_orig['h_keyboard']               = ( None                   , 'Keyboard Shortcuts keys')
kbkey_orig['gr_n']                     = ( ord(']')               , 'Go to to next group')
kbkey_orig['gr_p']                     = ( ord('[')               , 'Go to to previous group')
kbkey_orig['free_keys']                = ( ord('f')               , 'Show Free keys')
kbkey_orig['undo_key']                 = ( ord('x')               , 'Revert item to saved value')
kbkey_orig['conflict_switch']          = ( ord('0')               , 'Switch between cocnflicting items')


# ! Window Keys:
kbkey_orig['h_windows']                = ( None                   , 'Windows keys')
kbkey_orig['F7']                       = ( curses.KEY_F7          , 'Remove old istallation files')
kbkey_orig['F8']                       = ( curses.KEY_F8          , 'Media Players management')
kbkey_orig['F9']                       = ( curses.KEY_F9          , 'Show EXE location')
kbkey_orig['F10']                      = ( curses.KEY_F10         , 'Uninstall PyRadio')

# ! TTS Keys
kbkey_orig['h_tts']                 = ( None                      , 'TTS Specific keys')
kbkey_orig['tts_test']              = ( ord('T')                  , 'Test TTS settings (config window)')
kbkey_orig['tts_help']              = ( ord('t')                  , 'Speak additional info')

if platform.system().lower().startswith('win'):
    kbkey_orig['rb_p_first']           = ( curses.KEY_F1          , 'Go to first search results page')
    kbkey_orig['rb_p_next']            = ( curses.KEY_F3          , 'Go to next search results page')
    kbkey_orig['rb_p_prev']            = ( curses.KEY_F2          , 'Go to previous search results page')

# keys are the same as the headers of kbkey_orig
conflicts = {}
for a_key in kbkey_orig:
    if kbkey_orig[a_key][0] is None:
        header = a_key
        conflicts[header] = []
    else:
        conflicts[header].append(a_key)

''' this is the working dict
    it is a  deep copy of the original
    done this way for qick access
'''
def populate_dict():
    for key, value in kbkey_orig.items():
        if value[0]:
            kbkey[key] = value[0]
    return kbkey

kbkey = {}
kbkey = populate_dict()

# localized keys
lkbkey = None

def check_localized(char, k_list, return_key=False):
    # global kb_letter
    # global lkbkey
    # logger.error('\n\n')
    # logger.error(f'{k_list = }')
    # logger.error(f'{kb_letter = }')
    # logger.error(f'{lkbkey = }')
    if platform.system().lower().startswith('win'):
        if char and isinstance(k_list, Iterable) and lkbkey is not None:
            # logger.error(f'{char = }')
            # logger.error(f'{chr(char) = }')
            for n in k_list:
                # logger.error(f'k_list: {n = }')
                # logger.error(f'*** { kb_letter =  }')
                # logger.error(f'*** { lkbkey =  }')
                key = chr(char)
                # logger.error(f'*** { key =  }')
                if char in lkbkey:
                    if ord(lkbkey[char]) == n:
                        if logger.isEnabledFor(logging.DEBUG):
                            logger.debug('localized char found: {0} => "{1}"'.format(char, lkbkey[char]))
                        # logger.error('\n\n')
                        if return_key:
                            # logger.info(f'returning {n}')
                            return n
                        # logger.info('returning True')
                        return True
                else:
                    # logger.error('char not in lkbkey')
                    continue
    else:
        if char and kb_letter and isinstance(k_list, Iterable) and lkbkey is not None:
            # logger.error(f'{char = }')
            # logger.error(f'{chr(char) = }')
            for n in k_list:
                # logger.error(f'k_list: {n = }')
                # logger.error(f'k_list: {chr(n) = }')
                # logger.error(f'*** { kb_letter =  }')
                # logger.error(f'*** { lkbkey =  }')
                if kb_letter in lkbkey:
                    x = lkbkey[kb_letter]
                    # logger.error(f'{x = }')
                    if x == chr(n):
                        if logger.isEnabledFor(logging.DEBUG):
                            logger.debug('localized char found: "{0}" => "{1}"'.format(kb_letter, x))
                        # logger.error('\n\n')
                        if return_key:
                            # logger.info(f'returning {n}')
                            return n
                        # logger.info('returning True')
                        return True
                else:
                    # logger.error('kb_letter not in lkbkey')
                    continue
    # logger.error('\n\n')
    if return_key:
        # logger.info('returning None')
        return None
    # logger.info('returning False')
    return False

def set_lkbkey(adict):
    global lkbkey
    lkbkey = adict

def get_lkbkey():
    global lkbkey
    return lkbkey

def set_kbkey(a_key, value):
    ''' update kbkey dict from other modules '''
    global kbkey  # This line comes after using kbkey
    kbkey[a_key] = value  # Attempting to use kbkey before declaring it as global

curses_function_keys_dict = {
    curses.KEY_F1: 'F1',
    curses.KEY_F2: 'F2',
    curses.KEY_F3: 'F3',
    curses.KEY_F4: 'F4',
    curses.KEY_F5: 'F5',
    curses.KEY_F6: 'F6',
    curses.KEY_F7: 'F7',
    curses.KEY_F8: 'F8',
    curses.KEY_F9: 'F9',
    curses.KEY_F10: 'F10',
}

curses_ascii_dict = {
    curses.ascii.NUL: '^@',
    curses.ascii.SOH: '^A',
    curses.ascii.STX: '^B',
    curses.ascii.ETX: '^C',
    curses.ascii.EOT: '^D',
    curses.ascii.ENQ: '^E',
    curses.ascii.ACK: '^F',
    curses.ascii.BEL: '^G',
    curses.ascii.BS:  '^H',
    curses.ascii.TAB: '^I',
    curses.ascii.LF:  '^J',
    curses.ascii.VT:  '^K',
    curses.ascii.FF:  '^L',
    curses.ascii.CR:  '^M',
    curses.ascii.SO:  '^N',
    curses.ascii.SI:  '^O',
    curses.ascii.DLE: '^P',
    curses.ascii.DC1: '^Q',
    curses.ascii.DC2: '^R',
    curses.ascii.DC3: '^S',
    curses.ascii.DC4: '^T',
    curses.ascii.NAK: '^U',
    curses.ascii.SYN: '^V',
    curses.ascii.ETB: '^W',
    curses.ascii.CAN: '^X',
    curses.ascii.EM:  '^Y',
    curses.ascii.SUB: '^Z',
    curses.ascii.ESC: '^[',
    curses.KEY_F1: 'F1',
    curses.KEY_F2: 'F2',
    curses.KEY_F3: 'F3',
    curses.KEY_F4: 'F4',
    curses.KEY_F5: 'F5',
    curses.KEY_F6: 'F6',
    curses.KEY_F7: 'F7',
    curses.KEY_F8: 'F8',
    curses.KEY_F9: 'F9',
    curses.KEY_F10: 'F10',
}

def read_keyboard_shortcuts(file_path, reset=False):
    global kbkey  # Declare kbkey as global since we're reassigning it
    if reset:
        kbkey = populate_dict()  # Reassign kbkey with a new OrderedDict
    data = None
    try:
        with open(file_path, 'r', encoding='utf-8', errors='ignore') as json_file:
            data = json.load(json_file)
    except (FileNotFoundError, json.JSONDecodeError, TypeError, IOError):
        pass
    if data is not None:
        for n in data.keys():
            kbkey[n] = data[n]  # Modify the existing kbkey

def read_localized_keyboard(localize, data_dir):
    ''' read localized file from disk

        localize
            the name of the language to be used
            the file will be localize + .json
        data_dir
            the package's data dir

        Finally populate global lbkey from the data dict
    '''
    # global lkbkey
    error = False
    data = None

    if localize is None or localize == 'english':
        error = True
    else:
        user_file = join(data_dir, 'lkb_' + localize + '.json')
        package_res = files("pyradio").joinpath("keyboard", 'lkb_' + localize + '.json')

        if exists(user_file):
            try:
                with open(user_file, 'r', encoding='utf-8', errors='ignore') as json_file:
                    data = json.load(json_file)
            except (json.JSONDecodeError, TypeError, IOError):
                error = True
        elif package_res.is_file():  # package resource
            with as_file(package_res) as tmp_path:
                try:
                    with open(tmp_path, 'r', encoding='utf-8', errors='ignore') as json_file:
                        data = json.load(json_file)
                except (json.JSONDecodeError, TypeError, IOError):
                    error = True
        else:
            error = True

    if error:
        # keys = list(string.ascii_lowercase) + list(string.ascii_uppercase)
        # values = list(string.ascii_lowercase) + list(string.ascii_uppercase)
        # data = {keys[i]: values[i] for i in range(len(keys))}
        set_lkbkey({})
        return

    # Reverse the keys and values
    if platform.system().lower().startswith('win'):
        reversed_dict = {ord(value): key for key, value in data.items()}
    else:
        reversed_dict = {value: key for key, value in data.items()}

    # logger.error('\n\nsetting lkbkey 2\n{}\n\n'.format(reversed_dict))
    set_lkbkey(reversed_dict)

def to_str(akey):
    ''' convert kbkey keys to a string '''
    # Handle function keys explicitly
    if kbkey[akey] == ord(' '):
        return 'Space'
    elif kbkey[akey] == curses.KEY_F1:
        return "F1"
    elif kbkey[akey] == curses.KEY_F2:
        return "F2"
    elif kbkey[akey] == curses.KEY_F3:
        return "F3"
    elif kbkey[akey] == curses.KEY_F4:
        return "F4"
    elif kbkey[akey] == curses.KEY_F5:
        return "F5"
    elif kbkey[akey] == curses.KEY_F6:
        return "F6"
    elif kbkey[akey] == curses.KEY_F7:
        return "F7"
    elif kbkey[akey] == curses.KEY_F8:
        return "F8"
    elif kbkey[akey] == curses.KEY_F9:
        return "F9"
    elif kbkey[akey] == curses.KEY_F10:
        return "F10"
    adict = {
    'rec':          'Verital Line',
    'pause':        'Space',
    'gr':           '^G',
    'gr_next':      '^E',
    'gr_prev':      '^Y',
    'st_up':        '^U',
    'st_dn':        '^D',
    'p_next':       '^N',
    'p_prev':       '^P',
    }
    if akey in adict:
        return adict[akey]
    return chr(kbkey[akey])


def kb2str(msg):
    ''' convert a string to an appropriate
        form to be displayed to the user

        All keys in kbkey will be replaced with to_str result
        provided they are enclosed to {}
    '''
    def repl(match):
        key = match.group(1)
        if key in kbkey:
            return to_str(key)
        return match.group(0)

    if '{' not in msg:
        return msg

    msg = _KB2STR_PATTERN.sub(repl, msg)

    if msg == ' ':
        msg = 'Space'

    return msg

def kb2strL(msg):
    ''' convert a string to an appropriate
        form to be displayed to the user

        msg can contain {X}, X is y, Y, n, N, q
    '''
    for n in ('y', 'Y', 'n', 'N', 'q'):
        msg = msg.replace('{' + n +  '}', chr(kbkey[n]))
    return msg

def kb2chr(akey):
    ''' convert a kbkey key to a string (result of to_str) '''
    if akey in kbkey.keys():
        return to_str(akey)
    return ''

def ctrl_code_to_string(a_code):
    if a_code:
        if a_code in curses_ascii_dict:
            return curses_ascii_dict[a_code]
        char = chr(a_code)
        return char
    return ''

def ctrl_code_to_letter(a_code):
    if a_code in curses_ascii_dict:
        return curses_ascii_dict[a_code][-1]
    return ''

def ctrl_code_to_simple_code(a_code):
    code = ctrl_code_to_letter(a_code)
    if code:
        return ord(code.lower())
    return None

def letter_to_ctrl_code(letter):
    ''' gets a letter (for example "a")
        returns the key of "^{letter}" (for example "^A")
        in curses_ascii_dict, or None for "^S", "^Z", and "^C" on Linux/macOS
        but accepts them on Windows.
    '''

    # Normalize to uppercase
    letter = letter.upper()

    # Calculate the ASCII value of the letter
    # ascii_value = ord(letter)

    # Check if it's a valid letter (A-Z)
    if 'A' <= letter <= 'Z':
        # Calculate the control character
        # control_char = curses.ascii.NUL + (ascii_value - ord('A') + 1)

        # Check for "^S", "^Z", and "^C" based on OS
        if platform.system() in ['Linux', 'Darwin']:  # Darwin is macOS
            if value := f'^{letter}' in ['^S', '^Z', '^C']:
                return None

        # Find and return the key corresponding to the control character
        for key, value in curses_ascii_dict.items():
            if value == f'^{letter}':
                return key

    # Return None if not a valid letter or not found
    return None

def is_valid_char(char, win):
    """
    Checks if the character c is a valid ASCII character or a control combination.
    If c is invalid, it will read from the window to clear the input buffer.

    Parameters:
        c (int): The character code obtained from getch().
        win (curses.window): The curses window object for further input if needed.

    Returns:
        bool: True if c is valid, False otherwise.
    """
    if char in (9, ord('\t')):
        return False
    if (65 <= char <= 90) or (97 <= char <= 122) or (1 <= char <= 47) or \
            char in (
                ord('='), ord('.'), ord('+'),
                ord('`'), ord('-'),
                ord('1'), ord('2'), ord('3'),
                ord('4'), ord('5'), ord('6'),
                ord('7'), ord('8'), ord('9'),
                curses.KEY_F1,
                curses.KEY_F2,
                curses.KEY_F3,
                curses.KEY_F4,
                curses.KEY_F5,
                curses.KEY_F6,
                curses.KEY_F7,
                curses.KEY_F8,
                curses.KEY_F9,
                curses.KEY_F10,
            ):
        return True
    return False

def is_invalid_key(key):
    """
    Check if the pressed key is a special key (like HOME, END, PgUp, etc.).

    Args:
        key (int): The key code returned by curses.getch().

    Returns:
        bool: True if it's a special key, False otherwise.
    """
    # Check for special keys defined in curses
    special_keys = [
        curses.KEY_HOME,
        curses.KEY_END,
        curses.KEY_PPAGE,  # Page Up
        curses.KEY_NPAGE,  # Page Down
        curses.KEY_LEFT,
        curses.KEY_RIGHT,
        curses.KEY_UP,
        curses.KEY_DOWN,
        curses.KEY_IC,     # Insert
        curses.KEY_DC,     # Delete
        curses.KEY_BACKSPACE,
        # Add more keys as necessary
    ]
    if key in (
        curses.KEY_F1,
        curses.KEY_F2,
        curses.KEY_F3,
        curses.KEY_F4,
        curses.KEY_F5,
        curses.KEY_F6,
        curses.KEY_F7,
        curses.KEY_F8,
        curses.KEY_F9,
        curses.KEY_F10,
    ):
        logger.error('Key is F-[1-10]')
        return False

    # Check if the key is in the list of special keys or is greater than 255
    return key in special_keys or key > 255

def is_ctrl_key(key):
    """
    Check if the pressed key is a Ctrl-* key.

    Args:
        key (int): The key code returned by curses.getch().

    Returns:
        bool: True if it's a Ctrl-* key, False otherwise.
    """
    # Ctrl+A to Ctrl+Z correspond to ASCII values 1 to 26
    return 0 <= key <= 26

def chk_key(char, key, win):
    logger.error(f'{lkbkey = }')
    for n in lkbkey:
        logger.error(f'{n} ({chr(n)}) : {lkbkey[n]} ({chr(lkbkey[n])})')
    logger.error(f'{key = }')
    logger.error(f'{chr(key) = }')
    logger.error(f'{char = }')
    logger.error(f'{chr(char) = }')
    if char == key:
        return True
    '''
    try:
        letter = get_unicode_and_cjk_char(None, char)
        logger.error(f'{letter = }')
        this_char = ord(letter)
        logger.error(f'{char(letter) = }')
    except:
        this_char = None
    '''
    letter = get_unicode_and_cjk_char(win, char)
    if letter is not None:
        logger.error(f'{letter = }')
        this_char = ord(letter)
        logger.error(f'{chr(this_char) = }')
        try:
            if this_char == lkbkey[key]:
                return True
        except IndexError:
            pass
    else:
        logger.error('letter is None')
    return False


def set_kb_letter(letter):
    global kb_letter
    try:
        if letter and letter.isprintable():  # Check if the letter is printable
            kb_letter = letter
            # logger.error(f'>>> {kb_letter = }')  # Log the printable letter
        else:
            kb_letter = ''  # Ignore non-printable characters
            # logger.error(f'>>> Ignored non-printable letter: {repr(letter)}')
    except AttributeError:
        kb_letter = ''
    if logger.isEnabledFor(logging.DEBUG) and kb_letter:
        logger.debug(f'setting {kb_letter = }')

def get_kb_letter():
    global kb_letter
    # logger.error(f'*** {kb_letter = }')
    return kb_letter

def set_kb_cjk(value):
    global kb_cjk
    kb_cjk = value

def get_unicode_and_cjk_char(win, char):
    # logger.error(f'{char = }')
    def _decode_string(data):
        """ convert bytes to string """
        encodings = ['utf-8', locale.getpreferredencoding(False), 'latin1']
        for enc in encodings:
            if enc:      # in case locale.getpreferredencoding returns None
                try:
                    return data.decode(enc)
                except UnicodeDecodeError:
                    continue
        # if nothing works, play safe
        return data.decode('latin1', errors='replace')

    def get_check_next_byte(win):
        # logger.error(f'{win = }')
        char = win.getch()
        enqueue_input(char)
        if 128 <= char <= 191:
            return char
        return None
        raise UnicodeError


    set_kb_cjk(False)
    # logger.info('reseting kb_letter')
    set_kb_letter('')
    # logger.error(f'all {win = }')
    bytes_data = []
    if char <= 127:
        ''' 1 byte '''
        bytes_data.append(char)
    #elif 194 <= char <= 223:
    elif 192 <= char <= 223:
        ''' 2 bytes_data '''
        bytes_data.append(char)
        bytes_data.append(get_check_next_byte(win))
    elif 224 <= char <= 239:
        ''' 3 bytes_data '''
        bytes_data.append(char)
        bytes_data.append(get_check_next_byte(win))
        bytes_data.append(get_check_next_byte(win))
    elif 240 <= char <= 244:
        ''' 4 bytes_data '''
        bytes_data.append(char)
        bytes_data.append(get_check_next_byte(win))
        bytes_data.append(get_check_next_byte(win))
        bytes_data.append(get_check_next_byte(win))
    ''' no zero byte allowed '''
    while 0 in bytes_data:
        bytes_data.remove(0)

    try:
        buf = bytearray(bytes_data)
    except (ValueError, TypeError):
        return None
    out = _decode_string(buf)
    if out:
        # logger.info('setting kb_letter')
        set_kb_letter(out)
        if is_wide(out) and not kb_cjk:
            set_kb_cjk(True)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('=== CJK editing is ON ===')
    else:
        out = None
        # logger.info('invalid kb_letter')
        set_kb_letter('')
        set_kb_cjk(False)
    # if logger.isEnabledFor(logging.DEBUG):
    #    logger.debug(f'setting {kb_letter = }')
    return out

def add_l10n_to_functions_dict(functions):
    # logger.error(local_functions)
    local_keys = get_lkbkey()

    if local_keys is None:
        return functions

    local_functions = functions.copy()
    new_keys = {}
    for key, value in local_functions.items():
        # logger.error(f'{key}: {value}')
        for lkey, lvalue in local_keys.items():
            if ord(lvalue) == key:
                # logger.error(f'    {lkey}: {lvalue}')
                # logger.error('      {} {}'.format(ord(lvalue), value))
                # logger.error('      {} {}'.format(ord(lkey), value))
                new_keys[ord(lkey)] = value
                break
    # logger.error('\n\nnew_keys = {}'.format(new_keys))
    for key, value in new_keys.items():
        local_functions[key] = value
    return local_functions

def remove_l10n_from_global_functions(global_functions, shortcut_names):
    ''' remove key - value from functions dict
        It actually removes any referenct to the function
        that an item in shortcut_names refers to

        Parameters
        ==========
        global_functions  :  dict       { [key code]: function }
        shortcut_names    :  tupple     ('t', 'open')

        Rerurns
        =======
        A modified copy of global_functions
    '''
    ret = {}

    if global_functions is not None:
        ret = dict(global_functions)  # shallow copy

        for a_shortcut_name in shortcut_names:
            if kbkey[a_shortcut_name] in ret:
                func_to_remove = ret[kbkey[a_shortcut_name]]

                # Find all keys that point to this function
                keys_to_remove = [k for k, v in ret.items() if v is func_to_remove]

                # Delete all these keys
                for k in keys_to_remove:
                    del ret[k]

    return ret


class LetterProvider:
    """
    A class to provide and manage letter sequences based on various predefined keyboard layouts.

    This class maintains an internal index to select the current layout and offers methods to
    navigate through available layouts, retrieve layout sequences, and add new custom layouts.
    """

    def __init__(self):
        # Predefined layouts
        self.index = 1
        self._layouts = {
            "Alphabetical": "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
            "QWERTY": "qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM",
            "AZERTY": "azertyuiopqsdfghjklmwxcvbnAZERTYUIOPQSDFGHJKLMWXCVBN",
            "QWERTZ": "qwertzuiopasdfghjklyxcvbnmQWERTZUIOPASDFGHJKLYXCVBNM",
            "DVORAK": "pyfgcrlaoeuidhtnsqjkxbmwvzPYFGCRLAOEUIDHTNSQJKXBMWVZ",
            "COLEMAK": "qwfpgjluyarstdhneiozxcvbkmQWFPGJLUYARSTDHNEIOZXCVBKM",
        }

    @property
    def layout_name(self):
        """
        Gets the name of the currently selected keyboard layout, determined by `self.index`.

        Returns:
            str: The name of the current layout.
        """
        layouts = list(self._layouts.keys())
        return layouts[self.index]

    @property
    def max_length(self):
        """
        Returns the maximum number of characters among all stored keyboard layouts.

        Returns:
            int: The length of the longest layout string.
        """
        return max(len(x) for x in self._layouts)

    def get_layout(self):
        """
        Retrieves the letter sequence for the currently selected keyboard layout, based on `self.index`.

        If the current layout name is not found for any reason, the alphabetical layout is returned as a fallback.

        Returns:
            str: The sequence of letters in the current layout.
        """
        layout_name = list(self._layouts.keys())[self.index]
        return self._layouts.get(layout_name, self._layouts["Alphabetical"])

    def get_next_layout(self):
        """
        Advances `self.index` to select the next available keyboard layout and returns its name.

        If the end of the list is reached, wraps around to the first layout.

        Returns:
            str: The name of the newly selected layout.
        """
        layouts = list(self._layouts.keys())
        self.index += 1
        if self.index >= len(layouts):
            self.index = 0
        return layouts[self.index]

    def get_previous_layout(self):
        """
        Moves `self.index` to select the previous keyboard layout and returns its name.

        If the beginning of the list is reached, wraps around to the last layout.

        Returns:
            str: The name of the newly selected layout.
        """
        layouts = list(self._layouts.keys())
        self.index -= 1
        if self.index < 0:
            self.index = len(layouts) - 1
        return layouts[self.index]

    def add_layout(self, layout_name, layout_string):
        """
        Adds a new keyboard layout to the provider.

        The layout must be exactly 52 characters long, typically consisting of:
        26 lowercase letters (a-z) followed by 26 uppercase letters (A-Z).

        Args:
            layout_name (str): The name of the new keyboard layout.
            layout_string (str): The 52-character sequence (26 lowercase + 26 uppercase) for the new layout.

        Returns:
            None

        Raises:
            ValueError: If `layout_string` is not exactly 52 characters long.
        """
        if len(layout_string) != 52:
            raise ValueError("Layout string must contain exactly 52 characters (26 lowercase + 26 uppercase).")
        self._layouts[layout_name] = layout_string

    def list_layouts(self):
        """
    Lists all currently available keyboard layouts by name.

    Returns:
        list of str: A list of layout names, in the order they were defined or added.
    """
        return list(self._layouts.keys())


class LetterDisplay:

    def __init__(self, parent, focused=False, start_line=0):
        self._num_letters_per_line = 0
        self._width = 0
        self._height = 0
        self._active_widget = 0
        self._editing = None
        self._focused = False
        self._letter_width = 7
        self._layout = "QWERTY"
        self._provider = LetterProvider()
        self._left_pad = 0

        self._win = None
        self._parent = parent
        self._focused = focused
        self._start_line = start_line


        self._layouts = self._provider.list_layouts()
        self._letters = self._provider.get_layout()
        # self._letters = list(string.ascii_lowercase + string.ascii_uppercase)
        # self._letters_dict = {letter: 'α' for letter in string.ascii_letters}
        self._letters_dict = {}

        self._selection = 0
        self.next_table = [0] * len(self._letters)
        self.prev_table = [0] * len(self._letters)
        self._create_win(parent)
        self._win.keypad(True)

    @property
    def Y(self):
        return self._start_line

    @Y.setter
    def Y(self, val):
        create_win = self._start_line != val
        self._start_line = val
        if create_win:
            self._create_win()
            self.show()

    @property
    def height(self):
        return self._height

    @property
    def focused(self):
        return self._focused

    @focused.setter
    def focused(self, val):
        self._focused = val

    @property
    def letters_dict(self):
        return self._letters_dict

    @letters_dict.setter
    def letters_dict(self, val):
        if val:
            self._letters_dict = {}
            for n in val:
                self._letters_dict[n] = val[n]

    @property
    def editing(self):
        return self._editing

    @editing.setter
    def editing(self, value):
        if value:
            self._active_widget = 1
            self._selection = 0
            self._editing = value
        else:
            self._editing = None
            if self._active_widget == 1:
                self._active_widget = 0
                self.show()
                return
            self._show_editing()
        self._win.refresh()

    @property
    def active_widget(self):
        return self._active_widget

    @active_widget.setter
    def active_widget(self, value):
        self._active_widget = value

    @property
    def selected_letter(self):
        return self._letters[self._selection]

    @property
    def selection(self):
        return self._selection

    @selection.setter
    def selection(self, value):
        self._selection = value

    @property
    def left_pad(self):
        return self._left_pad

    @property
    def active_width(self):
        return self._num_letters_per_line * self._letter_width

    def _show_editing(self):
        if self.editing:
            self._win.addstr(
                0,
                self._left_pad + (self._num_letters_per_line * self._letter_width - self._provider.max_length) // 2 + len('Lowercase Letters  '),
                '  [',
                curses.color_pair(5))
            self._win.addstr(self.editing, curses.color_pair(2))
            self._win.addstr(']', curses.color_pair(5))
        else:
            self._win.addstr(
                0,
                self._left_pad + (self._num_letters_per_line * self._letter_width - self._provider.max_length) // 2 + len('Lowercase Letters  '),
                ' ' * (self._width - (self._num_letters_per_line * self._letter_width - self._provider.max_length) // 2 - len('Lowercase Letters  ') - 2),
                curses.color_pair(5))

    def set_letter(self, a_letter):
        self._letters_dict[self.selected_letter] = a_letter
        if self._selection < 26:
            sel = self._selection + 26
            self._letters_dict[self._letters[sel]] = a_letter.upper()
        self._selection = (self._selection + 1) % len(self._letters)
        self.show()

    def _calculate_num_of_letters_per_line(self):
        # self._num_letters_per_line = (self._width - 2) // self._letter_width
        self._num_letters_per_line = min(13, (self._width - 2) // self._letter_width)
        self._left_pad = int (( self._width - 2 - (self._letter_width * self._num_letters_per_line) ) / 2)
        if self._left_pad < 3:
            self._left_pad = 0

    def _clear_parent(self, col):
        self._parent.bkgd(' ', curses.color_pair(col))
        self._parent.clear()
        self._parent.refresh()

    def _create_win(self, parent=None):
        if self._win is not None:
            self._win.clear()
        if parent is not None:
            self._parent = parent
        # self._clear_parent(2)
        self.calculate_window_size()
        self._win = self._parent.subwin(self._height, self._width, self._start_line, 1)
        self._update_navigation_tables()
        self._win.bkgd(' ', curses.color_pair(5))
        self._win.erase()
        self._win.refresh()

    def calculate_window_size(self):
        _, self._width = self._parent.getmaxyx()
        self._width -= 2
        """Calculate the dynamic height required for the window."""
        self._calculate_num_of_letters_per_line()

        # Lines for lowercase letters
        lowercase_lines = (26 + self._num_letters_per_line - 1) // self._num_letters_per_line

        # Lines for uppercase letters
        uppercase_lines = (26 + self._num_letters_per_line - 1) // self._num_letters_per_line

        # Total lines including captions and spacing
        self._height = lowercase_lines + uppercase_lines + 4 # 3 display + 2 captions + 1 line of spacing
        return self._height

    def _update_navigation_tables(self):
        """Calculate the dynamic height required for the window."""
        self._calculate_num_of_letters_per_line()
        # logger.debug(f'{self._num_letters_per_line = }')

        lowercase_indices = list(range(26))
        uppercase_indices = list(range(26, len(self._letters)))
        # logger.debug(f'{ lowercase_indices = }')
        # logger.debug(f'{uppercase_indices = }')

        lowercase_columns = [lowercase_indices[i::self._num_letters_per_line] for i in range(self._num_letters_per_line)]
        uppercase_columns = [uppercase_indices[i::self._num_letters_per_line] for i in range(self._num_letters_per_line)]
        # logger.debug(f'{lowercase_columns = }')
        # logger.debug(f'{uppercase_columns = }')

        combined_columns = []
        for i in range(self._num_letters_per_line):
            combined_columns.append((lowercase_columns[i] if i < len(lowercase_columns) else []) +
                                    (uppercase_columns[i] if i < len(uppercase_columns) else []))
        # logger.debug(f'{combined_columns = }')

        for col in combined_columns:
            for j, index in enumerate(col):
                self.next_table[index] = col[(j + 1) % len(col)]
                self.prev_table[index] = col[(j - 1) % len(col)]
        # logger.debug(f'{self.next_table = }')
        # logger.debug(f'{self.prev_table = }')

    def show(self, parent=None):
        if parent and parent != self._parent:
            self._create_win(parent)

        """Calculate the dynamic height required for the window."""
        self._calculate_num_of_letters_per_line()

        # print layout

        start_line_offset = 0
        highlight = curses.color_pair(4) if self._focused else curses.color_pair(5)
        self._win.addstr(
            start_line_offset,
            self._left_pad + (self._num_letters_per_line * self._letter_width - self._provider.max_length) // 2,
            "Lowercase Letters",
            highlight
        )
        self._win.addstr(start_line_offset, 1 + self._left_pad, "Display:", curses.color_pair(5))
        if self._focused:
            if self._active_widget == 0:
                self._win.addstr(' ' + self._layout + ' ', curses.color_pair(9))
                self._win.addstr(' ' * (self._provider.max_length - len(self._layout)), curses.color_pair(2))
            else:
                self._win.addstr(' ' + self._layout + ' ', curses.color_pair(2))
                self._win.addstr(' ' * (self._provider.max_length - len(self._layout) + 1), curses.color_pair(2))
        else:
            self._win.addstr(' ' + self._layout.ljust(self._provider.max_length + 1), curses.color_pair(2))
        for i in range(0, 26, self._num_letters_per_line):
            for j in range(self._num_letters_per_line):
                index = i + j
                if index < 26:
                    self._draw_letter(index, i // self._num_letters_per_line + start_line_offset + 1, self._left_pad + j * self._letter_width + 2)

        uppercase_start_line = start_line_offset + ((26 + self._num_letters_per_line - 1) // self._num_letters_per_line) + 2
        self._win.addstr(
            uppercase_start_line,
             self._left_pad + (self._num_letters_per_line * self._letter_width - self._provider.max_length) // 2,
            "Uppercase Letters",
            highlight
        )
        for i in range(26, len(self._letters), self._num_letters_per_line):
            for j in range(self._num_letters_per_line):
                index = i + j
                if index < len(self._letters):
                    self._draw_letter(index, (i - 26) // self._num_letters_per_line + uppercase_start_line + 1, self._left_pad + j * self._letter_width + 2)

        selected_letter = self._letters[self._selection]

        self._show_editing()

        self._win.refresh()

    def _draw_letter(self, index, line, col):
        if line < self._height - 1 and col < self._width - 1:
            letter = self._letters[index]
            dict_value = self._letters_dict[letter]
            self._win.addstr(line, col, f"{letter}:", curses.color_pair(5))
            if self._focused:
                if index == self._selection and self._active_widget == 1:
                    self._win.addstr(line, col + len(letter) + 1, "    ", curses.color_pair(9))
                    self._win.addstr(line, col + len(letter) + 2, dict_value, curses.color_pair(9))
                else:
                    self._win.addstr(line, col + len(letter) + 1, "    ", curses.color_pair(2))
                    self._win.addstr(line, col + len(letter) + 2, dict_value.ljust(self._letter_width-3), curses.color_pair(2))
            else:
                self._win.addstr(line, col + len(letter) + 1, ' ' + dict_value.ljust(self._letter_width-3), curses.color_pair(2))

    def _read_file(self, index):
        error = False
        data = {}
        keys = list(string.ascii_lowercase) + list(string.ascii_uppercase)
        if index < 2:
            error = True
        else:
            try:
                with open(self._files[index][1], 'r', encoding='utf-8', errors='ignore') as json_file:
                    data = json.load(json_file)
            except (FileNotFoundError, json.JSONDecodeError, TypeError, IOError):
                error = True
        if error:
            data = {keys[i]: '' for i in range(len(keys))}
        return data

    def focus_previous(self):
        ''' focus previous widget
            - if active widget is the "Display",
              return -1, to decrease focus on parent
            - if active widget is the letters display,
              adjust self._active_widget, and return 0
        '''
        if self._active_widget == 1:
            self._active_widget = 0
            return 0
        return -1

    def focus_next(self):
        ''' focus next widget
            - if active widget is the letters display,
              return 1, to increase focus on parent
            - if active widget is the "Display",
              adjust self._active_widget, and return 0
        '''
        if self._active_widget == 0:
            self._active_widget = 1
            return 0
        return 1

    def keypress(self, char):
        ''' LetterDisplay keypress
            Returns
                -1: cancel
                 0: continue
                 1: ok - result in self.letters_dict
                 2: help
        '''
        if self._active_widget == 0:
            if char == curses.KEY_LEFT and self._focused:
                self._layout = self._provider.get_previous_layout()
                self._letters = self._provider.get_layout()
            elif char == curses.KEY_RIGHT and self._focused:
                self._layout = self._provider.get_next_layout()
                self._letters = self._provider.get_layout()
        elif self._active_widget == 1:
            if char == curses.KEY_LEFT and self._focused:
                self._selection = (self._selection - 1) % len(self._letters)
            elif char in (curses.KEY_RIGHT, ord('\n')) and self._focused:
                self._selection = (self._selection + 1) % len(self._letters)
            elif char == curses.KEY_UP and self._focused:
                self._selection = self.prev_table[self._selection]
            elif char == curses.KEY_DOWN and self._focused:
                self._selection = self.next_table[self._selection]
            elif char == curses.KEY_HOME and self._focused:
                self._selection = 0
            elif char == curses.KEY_END and self._focused:
                self._selection = len(self._letters) - 1

        if char == ord('?'):
            return 2

        if char == ord('\x1b'):
            return -1

        self.show()
        return 0



if __name__ == '__main__':
    with open('/home/spiros/projects/my-gits/pyradio/pyradio/keyboard/classes.json', 'r', encoding='utf-8') as f:
        res = json.load(f)
    print('===> classes.py: res')
    print(res)
    out = []
    for n in res:
        out += res[n]

    print('\n\n===> In list: out')
    out = list(set(out))
    print(out)

    missing = []

    for n in kbkey_orig:
        # print(f'checking "{n}"')
        if n not in out and kbkey_orig[n][0]:
            missing.append(n)
            # print(f'  missing: "{n}"')

    print('\n\n===> unused shortcuts')
    print(missing)
    global_functions = {
        'tag',
        't_tag',
        'transp',
        'v_up1',
        'v_up2',
        'v_up3',
        'v_dn1',
        'v_dn2',
        'mute',
        's_vol',
        't_calc_col',
        'repaint',
        # ord('b'): None,
    }

    missing_after_global_functions = []
    for n in missing:
        if n not in global_functions:
            missing_after_global_functions.append(n)

    print('\n\n===> unused shortcuts after removing global_functions')
    print(missing_after_global_functions)


    extra = [
    'new_playlist',
    'rename_playlist',
    'open_remote_control',
    'open_dirs',
    'change_player',
    'hist_top',
    'buffer',
    'open_buffer',
    'last_playlist',
    'clear_reg',
    'clear_all_reg',
    'unnamed',
    'html_help',
    ]

    miss = []
    for n in missing_after_global_functions:
        if n not in extra:
            miss.append(n)

    print('\n\n===> unused shortcuts after removing h_extra section')
    print(miss)



# -*- coding: utf-8 -*-

def get_encodings():
    encoding_list = [
        ['ascii', '646, us-ascii', 'English'],
        ['big5', 'big5-tw, csbig5', 'Traditional Chinese'],
        ['big5hkscs', 'big5-hkscs, hkscs', 'Traditional Chinese'],
        ['cp037', 'IBM037, IBM039', 'English'],
        ['cp273', '273, IBM273, csIBM273', 'German'],
        ['cp424', 'EBCDIC-CP-HE, IBM424', 'Hebrew'],
        ['cp437', '437, IBM437', 'English'],
        ['cp500', 'EBCDIC-CP-BE, EBCDIC-CP-CH, IBM500', 'Western Europe'],
        ['cp720', '', 'Arabic'],
        ['cp737', '', 'Greek'],
        ['cp775', 'IBM775', 'Baltic languages'],
        ['cp850', '850, IBM850', 'Western Europe'],
        ['cp852', '852, IBM852', 'Central and Eastern Europe'],
        ['cp855', '855, IBM855', 'Bulgarian, Byelorussian, Macedonian, Russian, Serbian'],
        ['cp856', '', 'Hebrew'],
        ['cp857', '857, IBM857', 'Turkish'],
        ['cp858', '858, IBM858', 'Western Europe'],
        ['cp860', '860, IBM860', 'Portuguese'],
        ['cp861', '861, CP-IS, IBM861', 'Icelandic'],
        ['cp862', '862, IBM862', 'Hebrew'],
        ['cp863', '863, IBM863', 'Canadian'],
        ['cp864', 'IBM864', 'Arabic'],
        ['cp865', '865, IBM865', 'Danish, Norwegian'],
        ['cp866', '866, IBM866', 'Russian'],
        ['cp869', '869, CP-GR, IBM869', 'Greek'],
        ['cp874', '', 'Thai'],
        ['cp875', '', 'Greek'],
        ['cp932', '932, ms932, mskanji, ms-kanji', 'Japanese'],
        ['cp949', '949, ms949, uhc', 'Korean'],
        ['cp950', '950, ms950', 'Traditional Chinese'],
        ['cp1006', '', 'Urdu'],
        ['cp1026', 'ibm1026', 'Turkish'],
        ['cp1125', '1125, ibm1125, cp866u, ruscii', 'Ukrainian'],
        ['cp1140', 'ibm1140', 'Western Europe'],
        ['cp1250', 'windows-1250', 'Central and Eastern Europe'],
        ['cp1251', 'windows-1251', 'Bulgarian, Byelorussian, Macedonian, Russian, Serbian'],
        ['cp1252', 'windows-1252', 'Western Europe'],
        ['cp1253', 'windows-1253', 'Greek'],
        ['cp1254', 'windows-1254', 'Turkish'],
        ['cp1255', 'windows-1255', 'Hebrew'],
        ['cp1256', 'windows-1256', 'Arabic'],
        ['cp1257', 'windows-1257', 'Baltic languages'],
        ['cp1258', 'windows-1258', 'Vietnamese'],
        ['cp65001', 'Windows UTF-8', 'Windows only'],
        ['euc_jp', 'eucjp, ujis, u-jis', 'Japanese'],
        ['euc_jis_2004', 'jisx0213, eucjis2004', 'Japanese'],
        ['euc_jisx0213', 'eucjisx0213', 'Japanese'],
        ['euc_kr', 'euckr, korean, ksc5601, ks_c-5601, ks_c-5601-1987, ksx1001, ks_x-1001', 'Korean'],
        ['gb2312', 'chinese, csiso58gb231280, euc-cn, euccn, eucgb2312-cn, gb2312-1980, gb2312-80, iso-ir-58', 'Simplified Chinese'],
        ['gbk', '936, cp936, ms936', 'Unified Chinese'],
        ['gb18030', 'gb18030-2000', 'Unified Chinese'],
        ['hz', 'hzgb, hz-gb, hz-gb-2312', 'Simplified Chinese'],
        ['iso2022_jp', 'csiso2022jp, iso2022jp, iso-2022-jp', 'Japanese'],
        ['iso2022_jp_1', 'iso2022jp-1, iso-2022-jp-1', 'Japanese'],
        ['iso2022_jp_2', 'iso2022jp-2, iso-2022-jp-2', 'Japanese, Korean, Simplified Chinese, Western Europe, Greek'],
        ['iso2022_jp_2004', 'iso2022jp-2004, iso-2022-jp-2004', 'Japanese'],
        ['iso2022_jp_3', 'iso2022jp-3, iso-2022-jp-3', 'Japanese'],
        ['iso2022_jp_ext', 'iso2022jp-ext, iso-2022-jp-ext', 'Japanese'],
        ['iso2022_kr', 'csiso2022kr, iso2022kr, iso-2022-kr', 'Korean'],
        ['latin', 'iso-8859-1, iso8859-1, 8859, cp819, latin_1, latin1, L1', 'West Europe'],
        ['iso8859-2', 'iso-8859-2, latin2, L2', 'Central and Eastern Europe'],
        ['iso8859-3', 'iso-8859-3, latin3, L3', 'Esperanto, Maltese'],
        ['iso8859-4', 'iso-8859-4, latin4, L4', 'Baltic languages'],
        ['iso8859-5', 'iso-8859-5, cyrillic', 'Bulgarian, Byelorussian, Macedonian, Russian, Serbian'],
        ['iso8859-6', 'iso-8859-6, arabic', 'Arabic'],
        ['iso8859-7', 'iso-8859-7, greek, greek8', 'Greek'],
        ['iso8859-8', 'iso-8859-8, hebrew', 'Hebrew'],
        ['iso8859-9', 'iso-8859-9, latin5, L5', 'Turkish'],
        ['iso8859-10', 'iso-8859-10, latin6, L6', 'Nordic languages'],
        ['iso8859-11', 'iso-8859-11, thai', 'Thai languages'],
        ['iso8859-13', 'iso-8859-13, latin7, L7', 'Baltic languages'],
        ['iso8859-14', 'iso-8859-14, latin8, L8', 'Celtic languages'],
        ['iso8859-15', 'iso-8859-15, latin9, L9', 'Western Europe'],
        ['iso8859-16', 'iso-8859-16, latin10, L10', 'South-Eastern Europe'],
        ['johab', 'cp1361, ms1361', 'Korean'],
        ['koi8_r', '', 'Russian'],
        ['koi8_t', '', 'Tajik'],
        ['koi8_u', '', 'Ukrainian'],
        ['kz1048', 'kz_1048, strk1048_2002, rk1048', 'Kazakh'],
        ['mac_cyrillic', 'maccyrillic', 'Bulgarian, Byelorussian, Macedonian, Russian, Serbian'],
        ['mac_greek', 'macgreek', 'Greek'],
        ['mac_iceland', 'maciceland', 'Icelandic'],
        ['mac_latin2', 'maclatin2, maccentraleurope', 'Central and Eastern Europe'],
        ['mac_roman', 'macroman, macintosh', 'Western Europe'],
        ['mac_turkish', 'macturkish', 'Turkish'],
        ['ptcp154', 'csptcp154, pt154, cp154, cyrillic-asian', 'Kazakh'],
        ['shift_jis', 'csshiftjis, shiftjis, sjis, s_jis', 'Japanese'],
        ['shift_jis_2004', 'shiftjis2004, sjis_2004, sjis2004', 'Japanese'],
        ['shift_jisx0213', 'shiftjisx0213, sjisx0213, s_jisx0213', 'Japanese'],
        ['utf_32', 'U32, utf32', 'All languages'],
        ['utf_32_be', 'UTF-32BE', 'All languages'],
        ['utf_32_le', 'UTF-32LE', 'All languages'],
        ['utf_16', 'U16, utf16', 'All languages'],
        ['utf_16_be', 'UTF-16BE', 'All languages'],
        ['utf_16_le', 'UTF-16LE', 'All languages'],
        ['utf_7', 'utf-7, U7, unicode-1-1-utf-7', 'All languages'],
        ['utf-8', 'U8, UTF, utf_8', 'All languages'],
        ['utf_8_sig', '', 'All languages'],
    ]
    return encoding_list

